"""
Reka OpenTelemetry instrumentation utility functions
"""

import time

from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    response_as_dict,
    calculate_tbt,
    get_chat_model_cost,
    common_span_attributes,
    record_completion_metrics,
)
from openlit.semcov import SemanticConvention


def format_content(messages):
    """
    Process a list of messages to extract content.
    """

    formatted_messages = []
    for message in messages:
        role = message["role"]
        content = message["content"]

        if isinstance(content, list):
            content_str = ", ".join(
                f"{item['type']}: {item['text'] if 'text' in item else item['image_url']}"
                if "type" in item
                else f"text: {item['text']}"
                for item in content
            )
            formatted_messages.append(f"{role}: {content_str}")
        else:
            formatted_messages.append(f"{role}: {content}")

    return "\n".join(formatted_messages)


def common_chat_logic(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    is_stream,
):
    """
    Process chat request and generate Telemetry
    """

    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    prompt = format_content(scope._kwargs.get("messages", ""))
    request_model = scope._kwargs.get("model", "reka-core-20240501")

    cost = get_chat_model_cost(
        request_model, pricing_info, scope._input_tokens, scope._output_tokens
    )

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
        SemanticConvention.GEN_AI_SYSTEM_REKAAI,
        scope._server_address,
        scope._server_port,
        request_model,
        scope._response_model,
        environment,
        application_name,
        is_stream,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Response parameters
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason]
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_SEED, scope._kwargs.get("seed", "")
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS,
        scope._kwargs.get("max_tokens", -1),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_PRESENCE_PENALTY,
        scope._kwargs.get("presence_penalty", 0.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES, scope._kwargs.get("stop", [])
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TEMPERATURE,
        scope._kwargs.get("temperature", 0.4),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TOP_K, scope._kwargs.get("top_k", 1.0)
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TOP_P, scope._kwargs.get("top_p", 1.0)
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_OUTPUT_TYPE,
        "text" if isinstance(scope._llmresponse, str) else "json",
    )

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, scope._output_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE,
        scope._input_tokens + scope._output_tokens,
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Tools
    if scope._tools:
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_NAME, scope._tools.get("name", "")
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_CALL_ID, str(scope._tools.get("id", ""))
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_ARGS, str(scope._tools.get("parameters", ""))
        )

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse
        )

        # To be removed one the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_completion_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            SemanticConvention.GEN_AI_SYSTEM_REKAAI,
            scope._server_address,
            scope._server_port,
            request_model,
            scope._response_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            scope._input_tokens,
            scope._output_tokens,
            cost,
            scope._tbt,
            scope._ttft,
        )


def process_chat_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process chat request and generate Telemetry
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._llmresponse = " ".join(
        (choice.get("message", {}).get("content") or "")
        for choice in response_dict.get("responses", [])
    )
    scope._response_id = response_dict.get("id")
    scope._response_model = response_dict.get("model")
    scope._input_tokens = response_dict.get("usage").get("input_tokens")
    scope._output_tokens = response_dict.get("usage").get("output_tokens")
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs
    scope._finish_reason = str(response_dict.get("responses")[0].get("finish_reason"))
    if scope._kwargs.get("tools"):
        scope._tools = (
            response_dict.get("responses")[0].get("message").get("tool_calls")
        )
    else:
        scope._tools = None

    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=False,
    )

    return response



"""
OpenLIT LangChain Callback Handler for Hierarchical Span Creation
"""

import time
import json
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional
from uuid import UUID

from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.messages import BaseMessage
from langchain_core.outputs import LLMResult, ChatGeneration, Generation

from opentelemetry import context as context_api
from opentelemetry.trace import SpanKind, set_span_in_context, Status, StatusCode
from opentelemetry.trace.span import Span

from openlit.__helpers import (
    common_framework_span_attributes,
    handle_exception,
    get_chat_model_cost,
    general_tokens,
)
from openlit.semcov import SemanticConvention

# Enhanced Provider Mapping (inspired by OpenInference)
LANGCHAIN_PROVIDER_MAP = {
    "anthropic": "anthropic",
    "azure": "azure",
    "azure_ai": "azure",
    "azure_openai": "azure",
    "bedrock": "aws",
    "bedrock_converse": "aws",
    "cohere": "cohere",
    "deepseek": "deepseek",
    "fireworks": "fireworks",
    "google": "google",
    "google_anthropic_vertex": "google",
    "google_genai": "google",
    "google_vertexai": "google",
    "groq": "groq",
    "huggingface": "huggingface",
    "ibm": "ibm",
    "mistralai": "mistralai",
    "nvidia": "nvidia",
    "ollama": "ollama",
    "openai": "openai",
    "perplexity": "perplexity",
    "together": "together",
    "vertex": "google",
    "vertexai": "google",
    "xai": "xai",
}


class SpanHolder:  # pylint: disable=too-few-public-methods
    """Holds span information and tracks relationships"""

    def __init__(self, span: Span, start_time: float, context_token=None):
        self.span = span
        self.start_time = start_time
        self.children: List[UUID] = []
        self.workflow_name = ""
        self.entity_name = ""
        self.context_token = context_token  # For context restoration


class OpenLITLangChainCallbackHandler(BaseCallbackHandler):
    """
    Enhanced OpenLIT callback handler
    """

    def __init__(
        self,
        tracer,
        version,
        environment,
        application_name,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    ):
        super().__init__()
        self.tracer = tracer
        self.version = version
        self.environment = environment
        self.application_name = application_name
        self.pricing_info = pricing_info
        self.capture_message_content = capture_message_content
        self.metrics = metrics
        self.disable_metrics = disable_metrics

        # Track active spans by run_id
        self.spans: Dict[UUID, SpanHolder] = {}

        self.session_name = environment  # Map environment to session
        self.tags_enabled = True  # Enable tagging system
        self.events_enabled = True  # Enable events tracking
        self.performance_baselines = {}  # Store performance baselines
        self.error_classification_enabled = True  # Enable error classification

    # Required BaseCallbackHandler properties
    @property
    def raise_error(self) -> bool:
        """Should the handler raise errors instead of logging them."""
        return False

    @property
    def run_inline(self) -> bool:
        """Should the handler run inline with the main thread."""
        return True

    # Ignore flags - all set to False so we capture everything
    @property
    def ignore_llm(self) -> bool:
        return False

    @property
    def ignore_chain(self) -> bool:
        return False

    @property
    def ignore_agent(self) -> bool:
        return False

    @property
    def ignore_retriever(self) -> bool:
        return False

    @property
    def ignore_chat_model(self) -> bool:
        return False

    def _get_span_name(self, serialized: Dict[str, Any], operation_type: str) -> str:
        """Generate OpenLIT-style span names following our naming convention"""

        # Handle None serialized (common for RunnableSequence)
        if not serialized:
            return f"{operation_type} RunnableSequence"

        # Extract class name for component identification
        if "id" in serialized and serialized["id"]:
            component_name = serialized["id"][-1]  # Last part is usually the class name
        elif "name" in serialized:
            component_name = serialized["name"]
        else:
            component_name = "unknown"

        # Follow OpenLIT naming: {operation_type} {component_name}
        return f"{operation_type} {component_name}"

    def _create_span(
        self,
        run_id: UUID,
        parent_run_id: Optional[UUID],
        span_name: str,
        kind: SpanKind = SpanKind.CLIENT,
        model_name: Optional[str] = None,
    ) -> Span:
        """Create a span with proper parent-child relationship and set as active context"""

        # If we have a parent, create child span in parent context
        if parent_run_id and parent_run_id in self.spans:
            parent_span = self.spans[parent_run_id].span
            span = self.tracer.start_span(
                span_name, context=set_span_in_context(parent_span), kind=kind
            )
            # Track parent-child relationship
            self.spans[parent_run_id].children.append(run_id)
        else:
            # Create root span
            span = self.tracer.start_span(span_name, kind=kind)

        # Set this span as the active context so downstream instrumentations (like OpenAI)
        # will create child spans under it
        span_context = set_span_in_context(span)
        context_token = context_api.attach(span_context)

        # Store span with start time and context token
        start_time = time.time()
        self.spans[run_id] = SpanHolder(span, start_time, context_token)

        # Set common framework span attributes for consistency
        scope = type("GenericScope", (), {})()
        scope._span = span
        scope._start_time = start_time
        scope._end_time = None

        # Create mock instance with model name for common_framework_span_attributes
        mock_instance = None
        if model_name:
            mock_instance = type("MockInstance", (), {"model_name": model_name})()

        common_framework_span_attributes(
            scope,
            SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN,
            "localhost",  # Default server address for LangChain
            8080,  # Default port
            self.environment,
            self.application_name,
            self.version,
            span_name,
            mock_instance,
        )

        return span

    def _end_span(self, run_id: UUID) -> None:
        """End span and all its children, restore context"""
        if run_id not in self.spans:
            return

        span_holder = self.spans[run_id]
        span = span_holder.span

        # End all child spans first
        for child_id in span_holder.children:
            if child_id in self.spans:
                self._end_span(child_id)

        # Restore the previous context before ending span
        if span_holder.context_token:
            context_api.detach(span_holder.context_token)

        # Update end time for duration calculation
        end_time = time.time()
        duration = end_time - span_holder.start_time
        span.set_attribute(
            SemanticConvention.GEN_AI_CLIENT_OPERATION_DURATION, duration
        )

        # End this span
        span.set_status(Status(StatusCode.OK))
        span.end()

        # Clean up
        del self.spans[run_id]

    def _add_langsmith_events(
        self, span: Span, event_type: str, data: Optional[Dict] = None
    ):
        """Add LangSmith-style events to spans"""
        if not self.events_enabled:
            return

        try:
            event_data = {
                "timestamp": datetime.now(timezone.utc).isoformat(),
                "event_type": event_type,
                "session_name": self.session_name,
            }

            if data:
                event_data.update(data)

            span.add_event(f"langchain.{event_type}", event_data)
        except Exception:
            pass  # Graceful degradation

    def _classify_error(self, error: Exception) -> str:
        """Classify errors like LangSmith does"""
        error_type = type(error).__name__

        # LangSmith-style error classification
        if "rate" in str(error).lower() or "429" in str(error):
            return "RATE_LIMIT_ERROR"
        elif "timeout" in str(error).lower():
            return "TIMEOUT_ERROR"
        elif "auth" in str(error).lower() or "401" in str(error):
            return "AUTH_ERROR"
        elif "not found" in str(error).lower() or "404" in str(error):
            return "NOT_FOUND_ERROR"
        elif "connection" in str(error).lower():
            return "CONNECTION_ERROR"
        elif "validation" in str(error).lower():
            return "VALIDATION_ERROR"
        else:
            return f"GENERAL_ERROR_{error_type}"

    def _add_tags_from_context(self, span: Span, run_id: UUID, **kwargs):
        """Add LangSmith-style tags to spans"""
        if not self.tags_enabled:
            return

        try:
            tags = []

            # Auto-generate tags based on context
            if "model" in kwargs:
                tags.append(f"model:{kwargs['model']}")
            if "temperature" in kwargs:
                tags.append(f"temperature:{kwargs['temperature']}")
            if hasattr(kwargs.get("invocation_params", {}), "stream"):
                tags.append(
                    f"streaming:{kwargs['invocation_params'].get('stream', False)}"
                )

            # Add session tag
            tags.append(f"session:{self.session_name}")

            # Add environment tag
            tags.append(f"env:{self.environment}")

            if tags:
                span.set_attribute(SemanticConvention.GEN_AI_FRAMEWORK_TAGS, tags)

        except Exception:
            pass  # Graceful degradation

    def _track_performance_baseline(
        self, span: Span, operation_name: str, duration_ms: float
    ):
        """Track performance against baselines like LangSmith"""
        try:
            # Store baseline if not exists
            if operation_name not in self.performance_baselines:
                self.performance_baselines[operation_name] = {
                    "avg_duration": duration_ms,
                    "min_duration": duration_ms,
                    "max_duration": duration_ms,
                    "count": 1,
                }
            else:
                baseline = self.performance_baselines[operation_name]
                baseline["count"] += 1
                baseline["avg_duration"] = (
                    baseline["avg_duration"] * (baseline["count"] - 1) + duration_ms
                ) / baseline["count"]
                baseline["min_duration"] = min(baseline["min_duration"], duration_ms)
                baseline["max_duration"] = max(baseline["max_duration"], duration_ms)

            # Add performance comparison attributes
            baseline = self.performance_baselines[operation_name]
            span.set_attribute(
                SemanticConvention.GEN_AI_FRAMEWORK_PERFORMANCE_VS_BASELINE,
                duration_ms / baseline["avg_duration"],
            )
            span.set_attribute(
                SemanticConvention.GEN_AI_FRAMEWORK_PERFORMANCE_BASELINE_AVG,
                baseline["avg_duration"],
            )
            span.set_attribute(
                SemanticConvention.GEN_AI_FRAMEWORK_PERFORMANCE_BASELINE_PERCENTILE,
                _calculate_percentile(duration_ms, baseline),
            )

        except Exception:
            pass  # Graceful degradation

    # Enhanced callback methods with new features
    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when an LLM starts"""

        try:
            # Create workflow span for chain operations
            span_name = self._get_span_name(serialized, "workflow")
            span = self._create_span(
                run_id, parent_run_id, span_name, SpanKind.INTERNAL
            )

            # Set OpenLIT attributes
            span.set_attribute(
                SemanticConvention.GEN_AI_SYSTEM,
                SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN,
            )
            span.set_attribute(
                SemanticConvention.GEN_AI_OPERATION,
                SemanticConvention.GEN_AI_OPERATION_TYPE_FRAMEWORK,
            )

            # Set workflow-specific attributes
            if serialized and "id" in serialized and serialized["id"]:
                span.set_attribute(
                    SemanticConvention.GEN_AI_WORKFLOW_TYPE, serialized["id"][-1]
                )
            else:
                span.set_attribute(
                    SemanticConvention.GEN_AI_WORKFLOW_TYPE, "RunnableSequence"
                )

            # LangSmith-style serialized function capture
            self._capture_serialized_info(span, serialized)

            # Capture input if enabled (with safe JSON serialization)
            if self.capture_message_content:
                try:
                    input_str = json.dumps(prompts, default=str)[:1000]
                    span.set_attribute(
                        SemanticConvention.GEN_AI_WORKFLOW_INPUT, input_str
                    )
                except Exception:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_WORKFLOW_INPUT, str(prompts)[:1000]
                    )

            # Framework enhancements - use attributes only (not events for non-chat operations)

            self._add_tags_from_context(span, run_id, **kwargs)

            # Additional metadata is captured through semantic conventions above

            # Extract provider information (OpenInference-inspired enhancement)
            self._extract_provider_info(span, **kwargs)

        except Exception:
            # Graceful error handling to prevent callback system failure
            pass

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when an LLM call ends"""

        try:
            if run_id not in self.spans:
                return

            span_holder = self.spans[run_id]
            span = span_holder.span

            # Process LLM response with OpenLIT's business intelligence
            self._process_llm_response(span, response, run_id)

            # Duration is set in _end_span method

            # NEW: Add performance baseline tracking
            if span_holder:
                duration_ms = (time.time() - span_holder.start_time) * 1000
                self._track_performance_baseline(span, span.name, duration_ms)

            # Framework completion - use attributes only (not events for non-chat operations)

            self._end_span(run_id)

        except Exception:
            # Graceful error handling
            pass

    def on_llm_error(
        self,
        error: Exception,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when an LLM call ends with an error"""

        try:
            if run_id not in self.spans:
                return

            span_holder = self.spans[run_id]
            span = span_holder.span

            # NEW: Enhanced error classification and tracking
            # Framework error classification
            error_class = self._classify_error(error)
            span.set_attribute(
                SemanticConvention.GEN_AI_FRAMEWORK_ERROR_CLASS, error_class
            )
            span.set_attribute(
                SemanticConvention.GEN_AI_FRAMEWORK_ERROR_TYPE, type(error).__name__
            )
            span.set_attribute(
                SemanticConvention.GEN_AI_FRAMEWORK_ERROR_MESSAGE, str(error)
            )

            # Framework error - use attributes only (not events for non-chat operations)

            # Set error status
            span.set_status(Status(StatusCode.ERROR, str(error)))
            span.record_exception(error)

            self._end_span(run_id)

        except Exception:
            # Graceful error handling
            pass

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a chain (RunnableSequence, etc.) starts"""

        try:
            # Create workflow span for chain operations
            span_name = self._get_span_name(serialized, "workflow")
            span = self._create_span(
                run_id, parent_run_id, span_name, SpanKind.INTERNAL
            )

            # Set OpenLIT attributes
            span.set_attribute(
                SemanticConvention.GEN_AI_SYSTEM,
                SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN,
            )
            span.set_attribute(
                SemanticConvention.GEN_AI_OPERATION,
                SemanticConvention.GEN_AI_OPERATION_TYPE_FRAMEWORK,
            )

            # Set workflow-specific attributes
            if serialized and "id" in serialized and serialized["id"]:
                span.set_attribute(
                    SemanticConvention.GEN_AI_WORKFLOW_TYPE, serialized["id"][-1]
                )
            else:
                span.set_attribute(
                    SemanticConvention.GEN_AI_WORKFLOW_TYPE, "RunnableSequence"
                )

            # LangSmith-style serialized function capture
            self._capture_serialized_info(span, serialized)

            # Capture input if enabled (with safe JSON serialization)
            if self.capture_message_content:
                try:
                    input_str = json.dumps(inputs, default=str)[:1000]
                    span.set_attribute(
                        SemanticConvention.GEN_AI_WORKFLOW_INPUT, input_str
                    )
                except Exception:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_WORKFLOW_INPUT, str(inputs)[:1000]
                    )

            # Enhanced chain tracking - use attributes only (not events for non-chat operations)

            # Add chain-specific tags
            chain_tags = [f"chain_type:{serialized.get('id', ['unknown'])[-1]}"]
            if inputs:
                chain_tags.append(f"input_count:{len(inputs)}")
            span.set_attribute(SemanticConvention.GEN_AI_FRAMEWORK_TAGS, chain_tags)

        except Exception:
            # Graceful error handling to prevent callback system failure
            pass

    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a chain ends"""

        try:
            if run_id not in self.spans:
                return

            span = self.spans[run_id].span

            # Capture output if enabled (with safe JSON serialization)
            if self.capture_message_content:
                try:
                    output_str = json.dumps(outputs, default=str)[:1000]
                    span.set_attribute(
                        SemanticConvention.GEN_AI_WORKFLOW_OUTPUT, output_str
                    )
                except Exception:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_WORKFLOW_OUTPUT, str(outputs)[:1000]
                    )

            # Duration is set in _end_span method

            self._end_span(run_id)

        except Exception:
            # Graceful error handling
            pass

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a chat model (ChatOpenAI, etc.) starts"""

        try:
            # Extract model name from multiple sources with fallback chain
            model_name = "unknown"

            # DEBUG: Log serialized data to understand structure
            # This helps us improve model extraction logic
            # print(f"DEBUG: serialized={serialized}")
            # print(f"DEBUG: kwargs={kwargs}")

            # Try extracting from serialized kwargs first (most reliable)
            if (
                serialized
                and "kwargs" in serialized
                and "model" in serialized["kwargs"]
            ):
                model_name = serialized["kwargs"]["model"]
            elif kwargs.get("model"):
                model_name = kwargs["model"]
            elif serialized:
                # Try extracting from different parts of serialized data
                # LangChain often stores model info in various locations
                if "model" in serialized:
                    model_name = serialized["model"]
                elif "model_name" in serialized:
                    model_name = serialized["model_name"]
                elif (
                    "name" in serialized and "model" in str(serialized["name"]).lower()
                ):
                    model_name = serialized["name"]
                elif "id" in serialized and serialized["id"]:
                    # Extract from class identifier - this gives us the framework being used
                    class_info = serialized["id"]
                    if isinstance(class_info, list) and len(class_info) > 0:
                        class_name = class_info[
                            -1
                        ]  # Last part is usually the class name
                        # Infer model based on LangChain provider class
                        if "chatopenai" in class_name.lower():
                            model_name = "gpt-3.5-turbo"  # Default OpenAI model
                        elif "chatanthropic" in class_name.lower():
                            model_name = "claude-3"
                        elif (
                            "chatgooglevertexai" in class_name.lower()
                            or "chatgoogleai" in class_name.lower()
                        ):
                            model_name = "gemini-pro"
                        elif "llama" in class_name.lower():
                            model_name = "llama-2"
                        else:
                            # Use the class name itself as model identifier
                            model_name = class_name

            # Create chat span with model name
            span_name = f"chat {model_name}"
            span = self._create_span(
                run_id, parent_run_id, span_name, SpanKind.CLIENT, model_name
            )

            # Set OpenLIT chat operation attributes
            span.set_attribute(
                SemanticConvention.GEN_AI_SYSTEM, "openai"
            )  # Most common
            span.set_attribute(
                SemanticConvention.GEN_AI_OPERATION,
                SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            )
            span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL, model_name)

            # Streaming detection
            is_streaming = kwargs.get("stream", False) or kwargs.get("streaming", False)
            span.set_attribute(
                SemanticConvention.GEN_AI_REQUEST_IS_STREAM, is_streaming
            )

            # Process messages for content capture and token counting
            if messages and len(messages) > 0:
                formatted_messages = self._format_messages(messages[0])

                if self.capture_message_content:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_CONTENT_PROMPT, formatted_messages
                    )

                # Calculate input tokens
                input_tokens = general_tokens(formatted_messages)
                span.set_attribute(
                    SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, input_tokens
                )

        except Exception:
            # Graceful error handling
            pass

    def _format_messages(self, messages: List[BaseMessage]) -> str:
        """Format LangChain messages for content capture"""
        formatted = []
        for message in messages:
            role = self._get_message_role(message)
            content = getattr(message, "content", str(message))
            formatted.append(f"{role}: {content}")
        return "\n".join(formatted)

    def _get_message_role(self, message: BaseMessage) -> str:
        """Extract role from LangChain message"""
        message_type = message.__class__.__name__.lower()
        if "human" in message_type:
            return "user"
        elif "ai" in message_type:
            return "assistant"
        elif "system" in message_type:
            return "system"
        elif "tool" in message_type:
            return "tool"
        else:
            return "user"

    def _process_llm_response(
        self, span: Span, response: LLMResult, run_id: UUID
    ) -> None:
        """Process LLM response with OpenLIT's comprehensive business intelligence"""

        try:
            # Extract response content
            if response.generations and len(response.generations) > 0:
                generation = response.generations[0][0]

                if isinstance(generation, ChatGeneration):
                    completion_content = generation.message.content
                elif isinstance(generation, Generation):
                    completion_content = generation.text
                else:
                    completion_content = str(generation)

                # Set completion content
                if self.capture_message_content:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_CONTENT_COMPLETION, completion_content
                    )

                # Calculate output tokens
                output_tokens = general_tokens(completion_content)
                span.set_attribute(
                    SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens
                )

            # Use enhanced token extraction method (inspired by OpenInference)
            self._extract_token_usage(span, response)

            # Extract additional LLM output information
            if hasattr(response, "llm_output") and response.llm_output:
                llm_output = response.llm_output

                # Model information
                if "model_name" in llm_output:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_RESPONSE_MODEL,
                        llm_output["model_name"],
                    )

            # Calculate cost using OpenLIT's cost tracking
            input_tokens = span.attributes.get(
                SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, 0
            )
            output_tokens = span.attributes.get(
                SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, 0
            )
            model_name = span.attributes.get(
                SemanticConvention.GEN_AI_REQUEST_MODEL, ""
            )

            if input_tokens and output_tokens and model_name:
                cost = get_chat_model_cost(
                    model_name, self.pricing_info, input_tokens, output_tokens
                )
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

        except Exception as e:
            handle_exception(span, e)

    def _extract_token_usage(self, span: Span, result: LLMResult) -> None:
        """Extract comprehensive token usage with OpenInference-style parsing"""
        try:
            # Try multiple token usage extraction patterns
            token_usage = None

            # Pattern 1: Standard non-streaming (OpenAI, most providers)
            if hasattr(result, "llm_output") and result.llm_output:
                token_usage = result.llm_output.get(
                    "token_usage"
                ) or result.llm_output.get("usage")

            # Pattern 2: Streaming outputs (when stream_usage=True)
            if not token_usage and result.generations:
                try:
                    first_gen = result.generations[0][0]
                    if hasattr(first_gen, "message") and hasattr(
                        first_gen.message, "kwargs"
                    ):
                        token_usage = first_gen.message.kwargs.get("usage_metadata")
                except (IndexError, AttributeError):
                    pass

            # Pattern 3: VertexAI-specific (generation_info.usage_metadata)
            if not token_usage and result.generations:
                try:
                    first_gen = result.generations[0][0]
                    if (
                        hasattr(first_gen, "generation_info")
                        and first_gen.generation_info
                    ):
                        token_usage = first_gen.generation_info.get("usage_metadata")
                except (IndexError, AttributeError):
                    pass

            if token_usage:
                # Support multiple token field names from different providers
                input_tokens = (
                    token_usage.get("prompt_tokens")
                    or token_usage.get("input_tokens")  # Anthropic
                    or token_usage.get("prompt_token_count")
                )  # Gemini

                output_tokens = (
                    token_usage.get("completion_tokens")
                    or token_usage.get("output_tokens")  # Anthropic
                    or token_usage.get("candidates_token_count")
                )  # Gemini

                total_tokens = token_usage.get("total_tokens") or token_usage.get(
                    "total_token_count"
                )  # Gemini

                if input_tokens:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, input_tokens
                    )
                if output_tokens:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens
                    )
                if total_tokens:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens
                    )

                # Enhanced token details (OpenAI-specific)
                if details := token_usage.get("completion_tokens_details"):
                    if audio_tokens := details.get("audio_tokens"):
                        span.set_attribute(
                            SemanticConvention.GEN_AI_USAGE_COMPLETION_TOKENS_DETAILS_AUDIO,
                            audio_tokens,
                        )
                    if reasoning_tokens := details.get("reasoning_tokens"):
                        span.set_attribute(
                            SemanticConvention.GEN_AI_USAGE_COMPLETION_TOKENS_DETAILS_REASONING,
                            reasoning_tokens,
                        )

                if details := token_usage.get("prompt_tokens_details"):
                    if cached_tokens := details.get("cached_tokens"):
                        span.set_attribute(
                            SemanticConvention.GEN_AI_USAGE_PROMPT_TOKENS_DETAILS_CACHE_READ,
                            cached_tokens,
                        )
                    if audio_tokens := details.get("audio_tokens"):
                        span.set_attribute(
                            SemanticConvention.GEN_AI_USAGE_PROMPT_TOKENS_DETAILS_CACHE_READ,
                            audio_tokens,
                        )

                # Anthropic cache tokens
                if cache_read := token_usage.get("cache_read_input_tokens"):
                    span.set_attribute(
                        SemanticConvention.GEN_AI_USAGE_PROMPT_TOKENS_DETAILS_CACHE_READ,
                        cache_read,
                    )
                if cache_write := token_usage.get("cache_creation_input_tokens"):
                    span.set_attribute(
                        SemanticConvention.GEN_AI_USAGE_PROMPT_TOKENS_DETAILS_CACHE_WRITE,
                        cache_write,
                    )

        except Exception as e:
            handle_exception(span, e)

    def _extract_provider_info(self, span: Span, **kwargs) -> None:
        """Extract provider information with OpenInference-style mapping"""
        try:
            # Extract from metadata if available
            if "metadata" in kwargs:
                metadata = kwargs["metadata"]
                if isinstance(metadata, dict) and "ls_provider" in metadata:
                    provider = metadata["ls_provider"].lower()
                    mapped_provider = LANGCHAIN_PROVIDER_MAP.get(provider, provider)
                    span.set_attribute(
                        SemanticConvention.GEN_AI_SYSTEM, mapped_provider
                    )
                    span.set_attribute(
                        SemanticConvention.GEN_AI_REQUEST_PROVIDER, mapped_provider
                    )
                    return

            # Extract from invocation parameters
            if "invocation_params" in kwargs:
                params = kwargs["invocation_params"]
                if isinstance(params, dict):
                    # Look for provider indicators in model names
                    model = params.get("model") or params.get("model_name", "")
                    if model:
                        if "gpt-" in model or "o1-" in model:
                            span.set_attribute(
                                SemanticConvention.GEN_AI_SYSTEM, "openai"
                            )
                        elif "claude-" in model:
                            span.set_attribute(
                                SemanticConvention.GEN_AI_SYSTEM, "anthropic"
                            )
                        elif "gemini-" in model or "bison-" in model:
                            span.set_attribute(
                                SemanticConvention.GEN_AI_SYSTEM, "google"
                            )
                        elif "mistral-" in model:
                            span.set_attribute(
                                SemanticConvention.GEN_AI_SYSTEM, "mistralai"
                            )

        except Exception as e:
            handle_exception(span, e)

    def _capture_serialized_info(self, span: Span, serialized: Dict[str, Any]) -> None:
        """Capture LangSmith-style serialized function information"""
        if not serialized:
            return

        try:
            # Capture function name (LangSmith enhancement)
            if "name" in serialized and serialized["name"]:
                span.set_attribute(
                    SemanticConvention.GEN_AI_SERIALIZED_NAME, serialized["name"]
                )

            # Capture function signature if available
            if "signature" in serialized and serialized["signature"]:
                span.set_attribute(
                    SemanticConvention.GEN_AI_SERIALIZED_SIGNATURE,
                    str(serialized["signature"])[:500],
                )

            # Capture docstring if available
            if "doc" in serialized and serialized["doc"]:
                span.set_attribute(
                    SemanticConvention.GEN_AI_SERIALIZED_DOC,
                    str(serialized["doc"])[:200],
                )

            # Capture module information
            if "id" in serialized and isinstance(serialized["id"], list):
                module_path = (
                    ".".join(serialized["id"][:-1]) if len(serialized["id"]) > 1 else ""
                )
                if module_path:
                    span.set_attribute(
                        SemanticConvention.GEN_AI_SERIALIZED_MODULE, module_path
                    )

        except Exception:
            # Graceful failure for serialized info capture
            pass

    def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a tool starts"""

        span_name = self._get_span_name(serialized, "tool")
        span = self._create_span(run_id, parent_run_id, span_name, SpanKind.CLIENT)

        span.set_attribute(
            SemanticConvention.GEN_AI_SYSTEM, SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN
        )
        span.set_attribute(
            SemanticConvention.GEN_AI_OPERATION,
            SemanticConvention.GEN_AI_OPERATION_TYPE_TOOLS,
        )

        if self.capture_message_content:
            span.set_attribute(SemanticConvention.GEN_AI_TOOL_INPUT, input_str[:1000])

    def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a tool ends"""

        if run_id not in self.spans:
            return

        span = self.spans[run_id].span

        if self.capture_message_content:
            span.set_attribute(SemanticConvention.GEN_AI_TOOL_OUTPUT, output[:1000])

        # Duration is set in _end_span method

        self._end_span(run_id)

    def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a retriever starts"""

        span_name = self._get_span_name(serialized, "retrieval")
        span = self._create_span(run_id, parent_run_id, span_name, SpanKind.CLIENT)

        span.set_attribute(
            SemanticConvention.GEN_AI_SYSTEM, SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN
        )
        span.set_attribute(
            SemanticConvention.GEN_AI_OPERATION,
            SemanticConvention.GEN_AI_OPERATION_TYPE_RETRIEVE,
        )

        if self.capture_message_content:
            span.set_attribute(SemanticConvention.GEN_AI_RETRIEVAL_QUERY, query[:1000])

    def on_retriever_end(
        self,
        documents,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a retriever ends"""

        if run_id not in self.spans:
            return

        span = self.spans[run_id].span

        # Document count
        doc_count = len(documents) if documents else 0
        span.set_attribute(
            SemanticConvention.GEN_AI_RETRIEVAL_DOCUMENT_COUNT, doc_count
        )

        # Sample document content
        if self.capture_message_content and documents:
            sample_docs = []
            for doc in documents[:3]:  # First 3 docs
                if hasattr(doc, "page_content"):
                    sample_docs.append(doc.page_content[:200])
                else:
                    sample_docs.append(str(doc)[:200])
            span.set_attribute(
                SemanticConvention.GEN_AI_RETRIEVAL_DOCUMENTS, "; ".join(sample_docs)
            )

        # Duration is set in _end_span method

        self._end_span(run_id)


def _calculate_percentile(value: float, baseline: Dict[str, Any]) -> float:
    """Calculate rough percentile based on min/max"""
    try:
        min_val = baseline["min_duration"]
        max_val = baseline["max_duration"]
        if max_val == min_val:
            return 50.0
        return ((value - min_val) / (max_val - min_val)) * 100
    except:
        return 50.0  # Default to median



"""
Module for monitoring OpenAI API calls.
"""

import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import handle_exception, set_server_address_and_port
from openlit.instrumentation.openai.utils import (
    process_chat_chunk,
    process_response_chunk,
    process_chat_response,
    process_streaming_chat_response,
    process_streaming_response_response,
    process_response_response,
    process_embedding_response,
    process_image_response,
    process_audio_response,
)
from openlit.semcov import SemanticConvention


def async_chat_completions(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for OpenAI async chat completions.
    """

    class TracedAsyncStream:
        """
        Wrapper for async streaming responses to collect telemetry.
        """

        def __init__(
            self,
            wrapped,
            span,
            span_name,
            kwargs,
            server_address,
            server_port,
        ):
            self.__wrapped__ = wrapped
            self._span = span
            self._span_name = span_name
            self._llmresponse = ""
            self._response_id = ""
            self._response_model = ""
            self._finish_reason = ""
            self._system_fingerprint = ""
            self._service_tier = "auto"
            self._tools = None
            self._kwargs = kwargs
            self._start_time = time.time()
            self._end_time = None
            self._timestamps = []
            self._ttft = 0
            self._tbt = 0
            self._server_address = server_address
            self._server_port = server_port

        async def __aenter__(self):
            await self.__wrapped__.__aenter__()
            return self

        async def __aexit__(self, exc_type, exc_value, traceback):
            await self.__wrapped__.__aexit__(exc_type, exc_value, traceback)

        def __aiter__(self):
            return self

        def __getattr__(self, name):
            """Delegate attribute access to the wrapped object."""
            return getattr(self.__wrapped__, name)

        async def __anext__(self):
            try:
                chunk = await self.__wrapped__.__anext__()
                process_chat_chunk(self, chunk)
                return chunk
            except StopAsyncIteration:
                try:
                    with self._span:
                        process_streaming_chat_response(
                            self,
                            pricing_info=pricing_info,
                            environment=environment,
                            application_name=application_name,
                            metrics=metrics,
                            capture_message_content=capture_message_content,
                            disable_metrics=disable_metrics,
                            version=version,
                        )
                except Exception as e:
                    handle_exception(self._span, e)
                raise

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the OpenAI async chat completions call.
        """

        streaming = kwargs.get("stream", False)
        server_address, server_port = set_server_address_and_port(
            instance, "api.openai.com", 443
        )
        request_model = kwargs.get("model", "gpt-4o")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}"

        if streaming:
            awaited_wrapped = await wrapped(*args, **kwargs)
            span = tracer.start_span(span_name, kind=SpanKind.CLIENT)

            return TracedAsyncStream(
                awaited_wrapped, span, span_name, kwargs, server_address, server_port
            )

        else:
            with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
                start_time = time.time()
                response = await wrapped(*args, **kwargs)

                try:
                    response = process_chat_response(
                        response=response,
                        request_model=request_model,
                        pricing_info=pricing_info,
                        server_port=server_port,
                        server_address=server_address,
                        environment=environment,
                        application_name=application_name,
                        metrics=metrics,
                        start_time=start_time,
                        span=span,
                        capture_message_content=capture_message_content,
                        disable_metrics=disable_metrics,
                        version=version,
                        **kwargs,
                    )

                except Exception as e:
                    handle_exception(span, e)

                return response

    return wrapper


def async_responses(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for OpenAI async responses API.
    """

    class TracedAsyncStream:
        """
        Wrapper for async streaming responses to collect telemetry.
        """

        def __init__(
            self,
            wrapped,
            span,
            span_name,
            kwargs,
            server_address,
            server_port,
        ):
            self.__wrapped__ = wrapped
            self._span = span
            self._span_name = span_name
            self._llmresponse = ""
            self._response_id = ""
            self._response_model = ""
            self._finish_reason = ""
            self._input_tokens = 0
            self._output_tokens = 0
            self._reasoning_tokens = 0
            self._operation_type = "responses"
            self._service_tier = "default"
            self._tools = None
            self._response_tools = None
            self._kwargs = kwargs
            self._start_time = time.time()
            self._end_time = None
            self._timestamps = []
            self._ttft = 0
            self._tbt = 0
            self._server_address = server_address
            self._server_port = server_port

        async def __aenter__(self):
            await self.__wrapped__.__aenter__()
            return self

        async def __aexit__(self, exc_type, exc_value, traceback):
            await self.__wrapped__.__aexit__(exc_type, exc_value, traceback)

        def __aiter__(self):
            return self

        def __getattr__(self, name):
            """Delegate attribute access to the wrapped object."""
            return getattr(self.__wrapped__, name)

        async def __anext__(self):
            try:
                chunk = await self.__wrapped__.__anext__()
                process_response_chunk(self, chunk)
                return chunk
            except StopAsyncIteration:
                try:
                    with self._span:
                        process_streaming_response_response(
                            self,
                            pricing_info=pricing_info,
                            environment=environment,
                            application_name=application_name,
                            metrics=metrics,
                            capture_message_content=capture_message_content,
                            disable_metrics=disable_metrics,
                            version=version,
                        )
                except Exception as e:
                    handle_exception(self._span, e)
                raise

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the OpenAI async responses API call.
        """

        streaming = kwargs.get("stream", False)
        server_address, server_port = set_server_address_and_port(
            instance, "api.openai.com", 443
        )
        request_model = kwargs.get("model", "gpt-4o")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}"

        if streaming:
            awaited_wrapped = await wrapped(*args, **kwargs)
            span = tracer.start_span(span_name, kind=SpanKind.CLIENT)

            return TracedAsyncStream(
                awaited_wrapped, span, span_name, kwargs, server_address, server_port
            )

        else:
            with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
                start_time = time.time()
                response = await wrapped(*args, **kwargs)

                try:
                    response = process_response_response(
                        response=response,
                        request_model=request_model,
                        pricing_info=pricing_info,
                        server_port=server_port,
                        server_address=server_address,
                        environment=environment,
                        application_name=application_name,
                        metrics=metrics,
                        start_time=start_time,
                        span=span,
                        capture_message_content=capture_message_content,
                        disable_metrics=disable_metrics,
                        version=version,
                        **kwargs,
                    )

                except Exception as e:
                    handle_exception(span, e)

                return response

    return wrapper


def async_chat_completions_parse(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for OpenAI async chat completions parse.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the OpenAI async chat completions parse call.
        """

        server_address, server_port = set_server_address_and_port(
            instance, "api.openai.com", 443
        )
        request_model = kwargs.get("model", "gpt-4o")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)

            try:
                response = process_chat_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper


def async_embedding(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for OpenAI async embeddings.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the OpenAI async embeddings call.
        """

        server_address, server_port = set_server_address_and_port(
            instance, "api.openai.com", 443
        )
        request_model = kwargs.get("model", "text-embedding-ada-002")

        span_name = (
            f"{SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING} {request_model}"
        )

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)

            try:
                response = process_embedding_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper


def async_image_generate(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for OpenAI async image generation.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the OpenAI async image generation call.
        """

        server_address, server_port = set_server_address_and_port(
            instance, "api.openai.com", 443
        )
        request_model = kwargs.get("model", "dall-e-2")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_IMAGE} {request_model}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)
            end_time = time.time()

            try:
                response = process_image_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    end_time=end_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper


def async_image_variations(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for OpenAI async image variations.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the OpenAI async image variations call.
        """

        server_address, server_port = set_server_address_and_port(
            instance, "api.openai.com", 443
        )
        request_model = kwargs.get("model", "dall-e-2")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_IMAGE} {request_model}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)
            end_time = time.time()

            try:
                response = process_image_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    end_time=end_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper


def async_audio_create(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for OpenAI async audio creation.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the OpenAI async audio creation call.
        """

        server_address, server_port = set_server_address_and_port(
            instance, "api.openai.com", 443
        )
        request_model = kwargs.get("model", "tts-1")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_AUDIO} {request_model}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)
            end_time = time.time()

            try:
                response = process_audio_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    end_time=end_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper



"""
OpenLIT LangChain Instrumentation - Callback-Based Hierarchical Spans
"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.langchain.callback_handler import (
    OpenLITLangChainCallbackHandler,
)

_instruments = ("langchain >= 0.1.20",)


class CallbackManagerWrapper:  # pylint: disable=too-few-public-methods
    """Wrapper to inject OpenLIT callback handler into LangChain's callback system"""

    def __init__(self, callback_handler: OpenLITLangChainCallbackHandler):
        self.callback_handler = callback_handler

    def __call__(self, wrapped, instance, args, kwargs):
        """Inject OpenLIT callback handler when BaseCallbackManager is initialized"""

        # Call original initialization
        wrapped(*args, **kwargs)

        # Check if our callback handler is already registered
        for handler in instance.inheritable_handlers:
            if isinstance(handler, type(self.callback_handler)):
                break
        else:
            # Add our callback handler to the manager
            instance.add_handler(self.callback_handler, True)


class LangChainInstrumentor(BaseInstrumentor):
    """
    OpenLIT LangChain instrumentor using callback-based hierarchical span creation.

    This approach hooks into LangChain's built-in callback system to create
    proper parent-child span relationships automatically, providing superior
    observability with OpenLIT's comprehensive business intelligence.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        version = importlib.metadata.version("langchain")
        environment = kwargs.get("environment", "default")
        application_name = kwargs.get("application_name", "default")
        tracer = kwargs.get("tracer")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        metrics = kwargs.get("metrics_dict")
        disable_metrics = kwargs.get("disable_metrics")

        # Create OpenLIT callback handler with all configuration
        openlit_callback_handler = OpenLITLangChainCallbackHandler(
            tracer=tracer,
            version=version,
            environment=environment,
            application_name=application_name,
            pricing_info=pricing_info,
            capture_message_content=capture_message_content,
            metrics=metrics,
            disable_metrics=disable_metrics,
        )

        # Hook into LangChain's callback system
        # This automatically provides hierarchical spans for:
        # - RunnableSequence (workflow spans)
        # - PromptTemplate (task spans)
        # - ChatOpenAI (chat spans)
        # - Tools, Retrievers, etc.
        try:
            wrap_function_wrapper(
                module="langchain_core.callbacks",
                name="BaseCallbackManager.__init__",
                wrapper=CallbackManagerWrapper(openlit_callback_handler),
            )
        except Exception:
            # Graceful degradation if callback system unavailable
            pass

        # === COMPETITIVE ADVANTAGE SUMMARY ===
        # Callback-based approach provides:
        # 1. ✅ Proper span hierarchy (like OpenLLMetry/OpenInference)
        # 2. ✅ Comprehensive business intelligence (cost, tokens, performance)
        # 3. ✅ Automatic operation coverage (no manual function wrapping)
        # 4. ✅ Native LangChain integration (uses built-in callback system)
        # 5. ✅ Superior observability depth and quality

        # Result: Best of both worlds - hierarchy + business intelligence

    def _uninstrument(self, **kwargs):
        """Remove instrumentation"""
        try:
            from opentelemetry.instrumentation.utils import unwrap

            unwrap("langchain_core.callbacks", "BaseCallbackManager.__init__")
        except Exception:
            pass



"""
Anthropic OpenTelemetry instrumentation utility functions
"""

import time

from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    calculate_ttft,
    response_as_dict,
    calculate_tbt,
    get_chat_model_cost,
    record_completion_metrics,
    common_span_attributes,
)
from openlit.semcov import SemanticConvention


def format_content(messages):
    """
    Format the messages into a string for span events.
    """

    if not messages:
        return ""

    formatted_messages = []
    for message in messages:
        if isinstance(message, dict):
            role = message.get("role", "user")
            content = message.get("content", "")
        else:
            # Handle Anthropic object format
            role = getattr(message, "role", "user")
            content = getattr(message, "content", "")

        if isinstance(content, list):
            # Handle structured content (e.g., text + images)
            text_parts = []
            for part in content:
                if isinstance(part, dict) and part.get("type") == "text":
                    text_parts.append(part.get("text", ""))
            content = " ".join(text_parts)
        elif not isinstance(content, str):
            content = str(content)

        formatted_messages.append(f"{role}: {content}")

    return "\n".join(formatted_messages)


def process_chunk(scope, chunk):
    """
    Process a chunk of response data and update state.
    """

    end_time = time.time()
    # Record the timestamp for the current chunk
    scope._timestamps.append(end_time)

    if len(scope._timestamps) == 1:
        # Calculate time to first chunk
        scope._ttft = calculate_ttft(scope._timestamps, scope._start_time)

    chunked = response_as_dict(chunk)

    # Collect message IDs and input token from events
    if chunked.get("type") == "message_start":
        scope._response_id = chunked.get("message").get("id")
        scope._input_tokens = chunked.get("message").get("usage").get("input_tokens")
        scope._response_model = chunked.get("message").get("model")
        scope._response_role = chunked.get("message").get("role")

    # Collect message IDs and aggregated response from events
    if chunked.get("type") == "content_block_delta":
        if chunked.get("delta").get("text"):
            scope._llmresponse += chunked.get("delta").get("text")
        elif chunked.get("delta").get("partial_json"):
            scope._tool_arguments += chunked.get("delta").get("partial_json")

    if chunked.get("type") == "content_block_start":
        if chunked.get("content_block").get("id"):
            scope._tool_id = chunked.get("content_block").get("id")
        if chunked.get("content_block").get("name"):
            scope._tool_name = chunked.get("content_block").get("name")

    # Collect output tokens and stop reason from events
    if chunked.get("type") == "message_delta":
        scope._output_tokens = chunked.get("usage").get("output_tokens")
        scope._finish_reason = chunked.get("delta").get("stop_reason")


def common_chat_logic(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    is_stream,
):
    """
    Process chat request and generate Telemetry
    """

    scope._end_time = time.time()
    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    formatted_messages = format_content(scope._kwargs.get("messages", []))
    request_model = scope._kwargs.get("model", "claude-3-5-sonnet-latest")

    cost = get_chat_model_cost(
        request_model, pricing_info, scope._input_tokens, scope._output_tokens
    )

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
        SemanticConvention.GEN_AI_SYSTEM_ANTHROPIC,
        scope._server_address,
        scope._server_port,
        request_model,
        scope._response_model,
        environment,
        application_name,
        is_stream,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Request parameters
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS,
        scope._kwargs.get("max_tokens", -1),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES,
        scope._kwargs.get("stop_sequences", []),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TEMPERATURE,
        scope._kwargs.get("temperature", 1.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TOP_K, scope._kwargs.get("top_k", 1.0)
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TOP_P, scope._kwargs.get("top_p", 1.0)
    )

    # Span Attributes for Response parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason]
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_OUTPUT_TYPE,
        "text" if isinstance(scope._llmresponse, str) else "json",
    )

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, scope._output_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE,
        scope._input_tokens + scope._output_tokens,
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Handle tool calls if present
    if scope._tool_calls:
        # Optimized tool handling - extract name, id, and arguments
        tool_name = scope._tool_calls.get("name", "")
        tool_id = scope._tool_calls.get("id", "")
        tool_args = scope._tool_calls.get("input", "")

        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_NAME, tool_name)
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_CALL_ID, tool_id)
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_ARGS, str(tool_args))

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_PROMPT, formatted_messages
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse
        )

        # To be removed once the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: formatted_messages,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics
    if not disable_metrics:
        record_completion_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            SemanticConvention.GEN_AI_SYSTEM_ANTHROPIC,
            scope._server_address,
            scope._server_port,
            request_model,
            scope._response_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            scope._input_tokens,
            scope._output_tokens,
            cost,
            scope._tbt,
            scope._ttft,
        )


def process_streaming_chat_response(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content=False,
    disable_metrics=False,
    version="",
):
    """
    Process streaming chat response and generate telemetry.
    """

    if scope._tool_id != "":
        scope._tool_calls = {
            "id": scope._tool_id,
            "name": scope._tool_name,
            "input": scope._tool_arguments,
        }

    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=True,
    )


def process_chat_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process non-streaming chat response and generate telemetry.
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    # pylint: disable = no-member
    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._llmresponse = response_dict.get("content", [{}])[0].get("text", "")
    scope._response_role = response_dict.get("role", "assistant")
    scope._input_tokens = response_dict.get("usage").get("input_tokens")
    scope._output_tokens = response_dict.get("usage").get("output_tokens")
    scope._response_model = response_dict.get("model", "")
    scope._finish_reason = response_dict.get("stop_reason", "")
    scope._response_id = response_dict.get("id", "")
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs

    # Handle tool calls if present
    content_blocks = response_dict.get("content", [])
    scope._tool_calls = None
    for block in content_blocks:
        if block.get("type") == "tool_use":
            scope._tool_calls = {
                "id": block.get("id", ""),
                "name": block.get("name", ""),
                "input": block.get("input", ""),
            }
            break

    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=False,
    )

    return response



"""Initializer of Auto Instrumentation of AWS Bedrock Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.bedrock.bedrock import converse, converse_stream

_instruments = ("boto3 >= 1.34.138",)


class BedrockInstrumentor(BaseInstrumentor):
    """
    An instrumentor for AWS Bedrock client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        version = importlib.metadata.version("boto3")
        environment = kwargs.get("environment", "default")
        application_name = kwargs.get("application_name", "default")
        tracer = kwargs.get("tracer")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        metrics = kwargs.get("metrics_dict")
        disable_metrics = kwargs.get("disable_metrics")

        # sync
        wrap_function_wrapper(
            "botocore.client",
            "ClientCreator.create_client",
            converse(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # streaming
        wrap_function_wrapper(
            "botocore.client",
            "ClientCreator.create_client",
            converse_stream(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

    def _uninstrument(self, **kwargs):
        pass



"""
Module for monitoring Pinecone async API calls.
"""

import time
from opentelemetry.trace import SpanKind
from opentelemetry import context as context_api
from openlit.__helpers import handle_exception
from openlit.instrumentation.pinecone.utils import (
    process_vectordb_response,
    DB_OPERATION_MAP,
    set_server_address_and_port,
)


def async_general_wrap(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for Pinecone async function calls.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the Pinecone async function call.
        """

        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
            return await wrapped(*args, **kwargs)

        # Get server address and port using the standard helper
        server_address, server_port = set_server_address_and_port(instance)

        db_operation = DB_OPERATION_MAP.get(gen_ai_endpoint, "unknown")
        if db_operation == "create_collection":
            namespace = kwargs.get("name") or (args[0] if args else "unknown")
        else:
            namespace = kwargs.get("namespace") or (args[0] if args else "unknown")
        span_name = f"{db_operation} {namespace}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)

            try:
                # Process response and generate telemetry
                response = process_vectordb_response(
                    response,
                    db_operation,
                    server_address,
                    server_port,
                    environment,
                    application_name,
                    metrics,
                    start_time,
                    span,
                    capture_message_content,
                    disable_metrics,
                    version,
                    instance,
                    args,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper



"""Initializer of Auto Instrumentation of Groq Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.groq.groq import chat
from openlit.instrumentation.groq.async_groq import async_chat

_instruments = ("groq >= 0.5.0",)


class GroqInstrumentor(BaseInstrumentor):
    """
    An instrumentor for Groq client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name", "default")
        environment = kwargs.get("environment", "default")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        disable_metrics = kwargs.get("disable_metrics")
        version = importlib.metadata.version("groq")

        # Chat completions
        wrap_function_wrapper(
            "groq.resources.chat.completions",
            "Completions.create",
            chat(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # Chat completions
        wrap_function_wrapper(
            "groq.resources.chat.completions",
            "AsyncCompletions.create",
            async_chat(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

    def _uninstrument(self, **kwargs):
        pass



"""
Module for monitoring async LangChain Community operations.
"""

import logging
from opentelemetry import trace
from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import handle_exception
from openlit.instrumentation.langchain_community.utils import process_general_response

# Initialize logger for LangChain Community instrumentation
logger = logging.getLogger(__name__)


def async_general_wrap(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for GenAI operations.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI operation call.
        """

        # Prepare server address and port
        server_address = "127.0.0.1"
        server_port = "80"

        # Get the parent span from the tracer
        with tracer.start_as_current_span(
            gen_ai_endpoint, kind=trace.SpanKind.CLIENT
        ) as span:
            try:
                # Call the original async function
                response = await wrapped(*args, **kwargs)

                # Process the response using the utility function
                response = process_general_response(
                    response,
                    gen_ai_endpoint,
                    server_port,
                    server_address,
                    environment,
                    application_name,
                    span,
                    version,
                )

                span.set_status(Status(StatusCode.OK))

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper



"""
OpenAI OpenTelemetry instrumentation utility functions
"""

import time

from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    calculate_ttft,
    response_as_dict,
    calculate_tbt,
    get_chat_model_cost,
    get_embed_model_cost,
    get_audio_model_cost,
    get_image_model_cost,
    general_tokens,
    record_completion_metrics,
    record_embedding_metrics,
    record_audio_metrics,
    record_image_metrics,
    common_span_attributes,
)
from openlit.semcov import SemanticConvention


def handle_not_given(value, default=None):
    """
    Handle OpenAI's NotGiven values and None values by converting them to appropriate defaults.
    """
    if hasattr(value, "__class__") and value.__class__.__name__ == "NotGiven":
        return default
    if value is None:
        return default
    return value


def format_content(messages):
    """
    Format the messages into a string for span events.
    Handles both chat completions format and responses API input format.
    """

    if not messages:
        return ""

    # Handle string input (simple case)
    if isinstance(messages, str):
        return messages

    # Handle list of messages
    formatted_messages = []

    for message in messages:
        try:
            role = message.get("role", "user") or message.role
            content = message.get("content", "") or message.content

        except:
            role = "user"
            content = str(messages)

        if isinstance(content, list):
            content_str_list = []
            for item in content:
                # Chat completions format
                if item.get("type") == "text":
                    content_str_list.append(f"text: {item.get('text', '')}")
                elif item.get("type") == "image_url" and not item.get(
                    "image_url", {}
                ).get("url", "").startswith("data:"):
                    content_str_list.append(f"image_url: {item['image_url']['url']}")

                # Responses API format
                elif item.get("type") == "input_text":
                    content_str_list.append(f"text: {item.get('text', '')}")
                elif item.get("type") == "input_image":
                    image_url = item.get("image_url", "")
                    if image_url and not image_url.startswith("data:"):
                        content_str_list.append(f"image_url: {image_url}")

            content_str = ", ".join(content_str_list)
            formatted_messages.append(f"{role}: {content_str}")
        else:
            formatted_messages.append(f"{role}: {content}")

    return "\n".join(formatted_messages)


def process_chat_chunk(scope, chunk):
    """
    Process a chunk of chat response data and update state.
    """

    end_time = time.time()
    scope._timestamps.append(end_time)

    if len(scope._timestamps) == 1:
        scope._ttft = calculate_ttft(scope._timestamps, scope._start_time)

    chunked = response_as_dict(chunk)

    # Extract content from chat completions
    if len(chunked.get("choices", [])) > 0 and "delta" in chunked.get("choices")[0]:
        delta = chunked.get("choices")[0]["delta"]
        content = delta.get("content")
        if content:
            scope._llmresponse += content

        # Handle tool calls in streaming - optimized
        delta_tools = delta.get("tool_calls")
        if delta_tools:
            scope._tools = scope._tools or []

            for tool in delta_tools:
                idx = tool.get("index", 0)

                # Extend list if needed
                scope._tools.extend([{}] * (idx + 1 - len(scope._tools)))

                if tool.get("id"):  # New tool (id exists)
                    func = tool.get("function", {})
                    scope._tools[idx] = {
                        "id": tool["id"],
                        "function": {
                            "name": func.get("name", ""),
                            "arguments": func.get("arguments", ""),
                        },
                        "type": tool.get("type", "function"),
                    }
                elif (
                    scope._tools[idx] and "function" in tool
                ):  # Append args (id is None)
                    scope._tools[idx]["function"]["arguments"] += tool["function"].get(
                        "arguments", ""
                    )

    # Extract metadata
    scope._response_id = chunked.get("id") or scope._response_id
    scope._response_model = chunked.get("model") or scope._response_model

    try:
        scope._finish_reason = (
            chunked.get("choices", [])[0].get("finish_reason") or scope._finish_reason
        )
    except (IndexError, AttributeError, TypeError):
        scope._finish_reason = "stop"

    scope._system_fingerprint = (
        chunked.get("system_fingerprint") or scope._system_fingerprint
    )
    scope._service_tier = chunked.get("service_tier") or scope._service_tier


def process_response_chunk(scope, chunk):
    """
    Process a chunk of response API data and update state.
    """

    end_time = time.time()
    scope._timestamps.append(end_time)

    if len(scope._timestamps) == 1:
        scope._ttft = calculate_ttft(scope._timestamps, scope._start_time)

    chunked = response_as_dict(chunk)

    # Extract content from responses API
    if chunked.get("type") == "response.output_text.delta":
        scope._llmresponse += chunked.get("delta", "")

    # Handle tool calls in streaming for responses API
    elif chunked.get("type") == "response.output_item.added":
        # New tool call item added
        if not hasattr(scope, "_response_tools") or scope._response_tools is None:
            scope._response_tools = []

        item = chunked.get("item", {})
        if item.get("type") == "function_call":
            scope._response_tools.append(
                {
                    "id": item.get("id", ""),
                    "call_id": item.get("call_id", ""),
                    "name": item.get("name", ""),
                    "type": item.get("type", "function_call"),
                    "arguments": item.get("arguments", ""),
                    "status": item.get("status", "in_progress"),
                }
            )

    elif chunked.get("type") == "response.function_call_arguments.delta":
        # Tool arguments being streamed
        if hasattr(scope, "_response_tools") and scope._response_tools:
            item_id = chunked.get("item_id", "")
            delta = chunked.get("delta", "")

            # Find the tool by item_id and append arguments
            for tool in scope._response_tools:
                if tool.get("id") == item_id:
                    tool["arguments"] += delta
                    break

    elif chunked.get("type") == "response.function_call_arguments.done":
        # Tool arguments complete
        if hasattr(scope, "_response_tools") and scope._response_tools:
            item_id = chunked.get("item_id", "")
            final_arguments = chunked.get("arguments", "")

            # Update the tool with final arguments
            for tool in scope._response_tools:
                if tool.get("id") == item_id:
                    tool["arguments"] = final_arguments
                    break

    elif chunked.get("type") == "response.output_item.done":
        # Tool call item complete
        if hasattr(scope, "_response_tools") and scope._response_tools:
            item = chunked.get("item", {})
            item_id = item.get("id", "")

            # Update the tool with final status and data
            for tool in scope._response_tools:
                if tool.get("id") == item_id:
                    tool.update(
                        {
                            "call_id": item.get("call_id", tool.get("call_id", "")),
                            "name": item.get("name", tool.get("name", "")),
                            "arguments": item.get(
                                "arguments", tool.get("arguments", "")
                            ),
                            "status": item.get("status", "completed"),
                        }
                    )
                    break

    elif chunked.get("type") == "response.completed":
        response_data = chunked.get("response", {})
        scope._response_id = response_data.get("id")
        scope._response_model = response_data.get("model")
        scope._finish_reason = response_data.get("status")

        usage = response_data.get("usage", {})
        scope._input_tokens = usage.get("input_tokens", 0)
        scope._output_tokens = usage.get("output_tokens", 0)

        # Handle reasoning tokens
        output_tokens_details = usage.get("output_tokens_details", {})
        scope._reasoning_tokens = output_tokens_details.get("reasoning_tokens", 0)


def common_response_logic(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    is_stream,
):
    """
    Process responses API request and generate Telemetry
    """

    scope._end_time = time.time()
    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    # For responses API, format input using the same function as chat completions
    input_data = scope._kwargs.get("input", "")
    prompt = format_content(input_data)
    request_model = scope._kwargs.get("model", "gpt-4o")

    # Calculate tokens and cost
    if hasattr(scope, "_input_tokens") and scope._input_tokens:
        input_tokens = scope._input_tokens
        output_tokens = scope._output_tokens
    else:
        input_tokens = general_tokens(prompt)
        output_tokens = general_tokens(scope._llmresponse)

    cost = get_chat_model_cost(request_model, pricing_info, input_tokens, output_tokens)

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
        SemanticConvention.GEN_AI_SYSTEM_OPENAI,
        scope._server_address,
        scope._server_port,
        request_model,
        scope._response_model,
        environment,
        application_name,
        is_stream,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Request parameters specific to responses API
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TEMPERATURE,
        handle_not_given(scope._kwargs.get("temperature"), 1.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TOP_P,
        handle_not_given(scope._kwargs.get("top_p"), 1.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS,
        handle_not_given(scope._kwargs.get("max_output_tokens"), -1),
    )

    # Reasoning parameters
    reasoning = scope._kwargs.get("reasoning", {})
    if reasoning:
        if reasoning.get("effort"):
            scope._span.set_attribute(
                "gen_ai.request.reasoning_effort", reasoning.get("effort")
            )

    # Responses API specific attributes
    if hasattr(scope, "_service_tier"):
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_SERVICE_TIER, scope._service_tier
        )

    # Span Attributes for Response parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason]
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_OUTPUT_TYPE, "text")

    # Span Attributes for Tools (responses API structure) - optimized
    if hasattr(scope, "_response_tools") and scope._response_tools:
        tools = (
            scope._response_tools
            if isinstance(scope._response_tools, list)
            else [scope._response_tools]
        )

        names, ids, args = (
            zip(
                *[
                    (
                        t.get("name", ""),
                        str(t.get("call_id", "")),  # Use call_id for responses API
                        str(t.get("arguments", "")),
                    )
                    for t in tools
                    if isinstance(t, dict) and t
                ]
            )
            if tools
            else ([], [], [])
        )

        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_NAME, ", ".join(filter(None, names))
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_CALL_ID, ", ".join(filter(None, ids))
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_ARGS, ", ".join(filter(None, args))
        )

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, input_tokens + output_tokens
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Reasoning tokens
    if hasattr(scope, "_reasoning_tokens") and scope._reasoning_tokens > 0:
        scope._span.set_attribute(
            "gen_ai.usage.reasoning_tokens", scope._reasoning_tokens
        )

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse
        )

        # To be removed once the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics
    if not disable_metrics:
        record_completion_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            SemanticConvention.GEN_AI_SYSTEM_OPENAI,
            scope._server_address,
            scope._server_port,
            request_model,
            scope._response_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            input_tokens,
            output_tokens,
            cost,
            scope._tbt,
            scope._ttft,
        )


def process_streaming_response_response(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content=False,
    disable_metrics=False,
    version="",
):
    """
    Process streaming responses API response and generate telemetry.
    """

    common_response_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=True,
    )


def process_response_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process non-streaming responses API response and generate telemetry.
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span

    # Extract content from responses API structure with reasoning support
    output = response_dict.get("output", [])
    scope._llmresponse = ""
    scope._response_tools = None

    if output:
        # Find the message item in the output array (might not be first if reasoning is present)
        message_item = None
        for item in output:
            if item.get("type") == "message":
                message_item = item
                break
            if item.get("type") == "function_call":
                # Handle tool call
                scope._response_tools = [
                    {
                        "id": item.get("id", ""),
                        "call_id": item.get("call_id", ""),
                        "name": item.get("name", ""),
                        "type": item.get("type", "function_call"),
                        "arguments": item.get("arguments", ""),
                        "status": item.get("status", ""),
                    }
                ]

        # Extract content from message item if found
        if message_item:
            content = message_item.get("content", [])
            if content and len(content) > 0:
                scope._llmresponse = content[0].get("text", "")

    scope._response_id = response_dict.get("id")
    scope._response_model = response_dict.get("model")

    # Handle token usage including reasoning tokens
    usage = response_dict.get("usage", {})
    scope._input_tokens = usage.get("input_tokens", 0)
    scope._output_tokens = usage.get("output_tokens", 0)

    output_tokens_details = usage.get("output_tokens_details", {})
    scope._reasoning_tokens = output_tokens_details.get("reasoning_tokens", 0)

    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs
    scope._service_tier = response_dict.get("service_tier", "default")
    scope._finish_reason = response_dict.get("status", "completed")

    common_response_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=False,
    )

    return response


def common_chat_logic(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    is_stream,
):
    """
    Process chat request and generate Telemetry
    """

    scope._end_time = time.time()
    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    # Format messages for chat operations
    if hasattr(scope, "_operation_type") and scope._operation_type == "responses":
        # Handle responses API input format using format_content
        input_data = scope._kwargs.get("input", "")
        prompt = format_content(input_data)
    else:
        # Handle standard chat format
        prompt = format_content(scope._kwargs.get("messages", []))

    request_model = scope._kwargs.get("model", "gpt-4o")

    # Calculate tokens and cost
    if hasattr(scope, "_input_tokens") and scope._input_tokens:
        input_tokens = scope._input_tokens
        output_tokens = scope._output_tokens
    else:
        input_tokens = general_tokens(prompt)
        output_tokens = general_tokens(scope._llmresponse)

    cost = get_chat_model_cost(request_model, pricing_info, input_tokens, output_tokens)

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
        SemanticConvention.GEN_AI_SYSTEM_OPENAI,
        scope._server_address,
        scope._server_port,
        request_model,
        scope._response_model,
        environment,
        application_name,
        is_stream,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Request parameters
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_SEED,
        str(handle_not_given(scope._kwargs.get("seed"), "")),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_FREQUENCY_PENALTY,
        handle_not_given(scope._kwargs.get("frequency_penalty"), 0.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS,
        handle_not_given(scope._kwargs.get("max_tokens"), -1),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_PRESENCE_PENALTY,
        handle_not_given(scope._kwargs.get("presence_penalty"), 0.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES,
        handle_not_given(scope._kwargs.get("stop"), []),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TEMPERATURE,
        handle_not_given(scope._kwargs.get("temperature"), 1.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TOP_P,
        handle_not_given(scope._kwargs.get("top_p"), 1.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_USER,
        handle_not_given(scope._kwargs.get("user"), ""),
    )

    # Span Attributes for Response parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason]
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_OUTPUT_TYPE,
        "text" if isinstance(scope._llmresponse, str) else "json",
    )

    # OpenAI-specific attributes
    if hasattr(scope, "_system_fingerprint") and scope._system_fingerprint:
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_RESPONSE_SYSTEM_FINGERPRINT,
            scope._system_fingerprint,
        )
    if hasattr(scope, "_service_tier") and scope._service_tier:
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_SERVICE_TIER, scope._service_tier
        )

    # Span Attributes for Tools - optimized
    if hasattr(scope, "_tools") and scope._tools:
        tools = scope._tools if isinstance(scope._tools, list) else [scope._tools]

        names, ids, args = (
            zip(
                *[
                    (
                        t.get("function", {}).get("name", ""),
                        str(t.get("id", "")),
                        str(t.get("function", {}).get("arguments", "")),
                    )
                    for t in tools
                    if isinstance(t, dict) and t
                ]
            )
            if tools
            else ([], [], [])
        )

        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_NAME, ", ".join(filter(None, names))
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_CALL_ID, ", ".join(filter(None, ids))
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_ARGS, ", ".join(filter(None, args))
        )

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, input_tokens + output_tokens
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse
        )

        # To be removed once the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics
    if not disable_metrics:
        record_completion_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            SemanticConvention.GEN_AI_SYSTEM_OPENAI,
            scope._server_address,
            scope._server_port,
            request_model,
            scope._response_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            input_tokens,
            output_tokens,
            cost,
            scope._tbt,
            scope._ttft,
        )


def process_streaming_chat_response(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content=False,
    disable_metrics=False,
    version="",
):
    """
    Process streaming chat response and generate telemetry.
    """

    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=True,
    )


def process_chat_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process non-streaming chat response and generate telemetry.
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._llmresponse = " ".join(
        (choice.get("message", {}).get("content") or "")
        for choice in response_dict.get("choices", [])
    )
    scope._response_id = response_dict.get("id")
    scope._response_model = response_dict.get("model")
    scope._input_tokens = response_dict.get("usage", {}).get("prompt_tokens", 0)
    scope._output_tokens = response_dict.get("usage", {}).get("completion_tokens", 0)
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs
    scope._system_fingerprint = response_dict.get("system_fingerprint", "")
    scope._service_tier = response_dict.get("service_tier", "auto")
    scope._finish_reason = (
        str(response_dict.get("choices", [])[0].get("finish_reason", ""))
        if response_dict.get("choices")
        else ""
    )

    # Handle operation type for responses API
    if kwargs.get("_operation_type") == "responses":
        scope._operation_type = "responses"

    # Handle tool calls
    if kwargs.get("tools"):
        scope._tools = (
            response_dict.get("choices", [{}])[0].get("message", {}).get("tool_calls")
        )
    else:
        scope._tools = None

    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=False,
    )

    return response


def common_embedding_logic(
    scope,
    request_model,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
):
    """
    Common logic for processing embedding operations.
    """

    # Calculate cost
    cost = get_embed_model_cost(request_model, pricing_info, scope._input_tokens)

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING,
        SemanticConvention.GEN_AI_SYSTEM_OPENAI,
        scope._server_address,
        scope._server_port,
        request_model,
        request_model,
        environment,
        application_name,
        False,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Request parameters
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_ENCODING_FORMATS,
        [handle_not_given(scope._kwargs.get("encoding_format"), "float")],
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_USER,
        handle_not_given(scope._kwargs.get("user"), ""),
    )

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, scope._input_tokens
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Content
    if capture_message_content:
        input_data = scope._kwargs.get("input", "")
        formatted_content = (
            format_content(input_data)
            if isinstance(input_data, (list, dict))
            else str(input_data)
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_PROMPT, formatted_content
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics
    if not disable_metrics:
        record_embedding_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING,
            SemanticConvention.GEN_AI_SYSTEM_OPENAI,
            scope._server_address,
            scope._server_port,
            request_model,
            request_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            scope._input_tokens,
            cost,
        )


def common_image_logic(
    scope,
    request_model,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
):
    """
    Common logic for processing image operations.
    """

    # Calculate cost
    cost = get_image_model_cost(
        request_model,
        pricing_info,
        scope._kwargs.get("size", "1024x1024"),
        scope._kwargs.get("quality", "standard"),
    )

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_IMAGE,
        SemanticConvention.GEN_AI_SYSTEM_OPENAI,
        scope._server_address,
        scope._server_port,
        request_model,
        request_model,
        environment,
        application_name,
        False,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Request parameters
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_IMAGE_SIZE,
        handle_not_given(scope._kwargs.get("size"), "1024x1024"),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_IMAGE_QUALITY,
        handle_not_given(scope._kwargs.get("quality"), "standard"),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_USER,
        handle_not_given(scope._kwargs.get("user"), ""),
    )

    # Extract response data
    response_dict = scope._response_dict
    images_data = response_dict.get("data", [])
    response_created = response_dict.get("created")
    response_size = response_dict.get("size")
    response_quality = response_dict.get("quality")
    response_output_format = response_dict.get("output_format")

    # Span Attributes for Response
    if response_created:
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_RESPONSE_ID, str(response_created)
        )

    # Process image data and collect URLs/base64 content
    if images_data:
        # Collect image URLs or base64 content
        image_contents = []

        for image in images_data:
            # Collect image content (URL or base64)
            if image.get("url"):
                image_contents.append(image["url"])
            elif image.get("b64_json"):
                # For base64, we typically dont want to store the full content in spans
                # Just indicate its base64 format
                image_contents.append("[base64_image_data]")

        # Set image response data using semantic conventions
        if image_contents:
            scope._span.set_attribute(
                SemanticConvention.GEN_AI_RESPONSE_IMAGE, image_contents
            )

    # Response-level attributes if different from request
    if response_size:
        scope._span.set_attribute("gen_ai.response.image_size", response_size)
    if response_quality:
        scope._span.set_attribute("gen_ai.response.image_quality", response_quality)
    if response_output_format:
        scope._span.set_attribute(
            "gen_ai.response.output_format", response_output_format
        )

    # Span Attributes for Cost
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Content
    if capture_message_content:
        # Always collect the original prompt
        prompt = scope._kwargs.get("prompt", "")
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)

        # Collect and set revised prompts if available
        if images_data:
            revised_prompts = []
            for image in images_data:
                if image.get("revised_prompt"):
                    revised_prompts.append(image["revised_prompt"])

            # Set revised prompts as span attribute if any were found
            if revised_prompts:
                scope._span.set_attribute(
                    SemanticConvention.GEN_AI_CONTENT_REVISED_PROMPT, revised_prompts
                )

            # Add revised prompt events for detailed tracking
            for i, image in enumerate(images_data):
                if image.get("revised_prompt"):
                    scope._span.add_event(
                        name=SemanticConvention.GEN_AI_CONTENT_REVISED_PROMPT,
                        attributes={
                            SemanticConvention.GEN_AI_CONTENT_REVISED_PROMPT: image[
                                "revised_prompt"
                            ],
                            "image_index": i,
                        },
                    )

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics
    if not disable_metrics:
        record_image_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_IMAGE,
            SemanticConvention.GEN_AI_SYSTEM_OPENAI,
            scope._server_address,
            scope._server_port,
            request_model,
            request_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            cost,
        )


def common_audio_logic(
    scope,
    request_model,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
):
    """
    Common logic for processing audio operations.
    """

    # Calculate cost
    input_text = scope._kwargs.get("input", "")
    cost = get_audio_model_cost(request_model, pricing_info, input_text)

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_AUDIO,
        SemanticConvention.GEN_AI_SYSTEM_OPENAI,
        scope._server_address,
        scope._server_port,
        request_model,
        request_model,
        environment,
        application_name,
        False,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Request parameters
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_AUDIO_VOICE,
        handle_not_given(scope._kwargs.get("voice"), "alloy"),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_AUDIO_RESPONSE_FORMAT,
        handle_not_given(scope._kwargs.get("response_format"), "mp3"),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_AUDIO_SPEED,
        handle_not_given(scope._kwargs.get("speed"), 1.0),
    )

    # Span Attributes for Cost
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Content
    if capture_message_content:
        input_text = scope._kwargs.get("input", "")
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, input_text)

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics
    if not disable_metrics:
        record_audio_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_AUDIO,
            SemanticConvention.GEN_AI_SYSTEM_OPENAI,
            scope._server_address,
            scope._server_port,
            request_model,
            request_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            cost,
        )


def process_audio_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    end_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process audio generation response and generate telemetry.
    """

    scope = type("GenericScope", (), {})()

    scope._start_time = start_time
    scope._end_time = end_time
    scope._span = span
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs

    common_audio_logic(
        scope,
        request_model,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
    )

    return response


def process_embedding_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process embedding response and generate telemetry.
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._input_tokens = response_dict.get("usage", {}).get("prompt_tokens", 0)
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs

    common_embedding_logic(
        scope,
        request_model,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
    )

    return response


def process_image_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    end_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process image generation response and generate telemetry.
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = end_time
    scope._span = span
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs
    scope._response_dict = response_dict

    common_image_logic(
        scope,
        request_model,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
    )

    return response



# pylint: disable=useless-return, bad-staticmethod-argument, disable=duplicate-code
"""
Initializer of Auto Instrumentation of Crawl4AI Functions.
Supports Crawl4AI 0.7.x with comprehensive operation coverage.
"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.crawl4ai.crawl4ai import general_wrap
from openlit.instrumentation.crawl4ai.async_crawl4ai import async_general_wrap

_instruments = ("crawl4ai >= 0.7.0",)


class Crawl4AIInstrumentor(BaseInstrumentor):
    """
    An instrumentor for Crawl4AI's client library.
    Supports comprehensive 0.7.x operations including async crawling,
    batch processing, deep crawling, and extraction strategies.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name", "default_application")
        environment = kwargs.get("environment", "default_environment")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        disable_metrics = kwargs.get("disable_metrics")

        try:
            version = importlib.metadata.version("crawl4ai")
        except importlib.metadata.PackageNotFoundError:
            version = "unknown"

        # Wrap AsyncWebCrawler.arun() - Primary single URL crawling
        try:
            wrap_function_wrapper(
                "crawl4ai.async_webcrawler",
                "AsyncWebCrawler.arun",
                async_general_wrap(
                    "crawl4ai.arun",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        # Wrap AsyncWebCrawler.arun_many() - Batch/streaming crawling (0.7.x)
        try:
            wrap_function_wrapper(
                "crawl4ai.async_webcrawler",
                "AsyncWebCrawler.arun_many",
                async_general_wrap(
                    "crawl4ai.arun_many",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        # Wrap legacy WebCrawler.run() for backward compatibility
        try:
            wrap_function_wrapper(
                "crawl4ai.web_crawler",
                "WebCrawler.run",
                general_wrap(
                    "crawl4ai.run",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        # Wrap deep crawling operations (0.7.x)
        try:
            wrap_function_wrapper(
                "crawl4ai.deep_crawling.strategies",
                "BFSDeepCrawlStrategy.arun",
                async_general_wrap(
                    "crawl4ai.deep_crawl.bfs",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.deep_crawling.strategies",
                "DFSDeepCrawlStrategy.arun",
                async_general_wrap(
                    "crawl4ai.deep_crawl.dfs",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.deep_crawling.strategies",
                "BestFirstCrawlingStrategy.arun",
                async_general_wrap(
                    "crawl4ai.deep_crawl.best_first",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        # Wrap LLM and Extraction Strategy Operations (0.7.x)
        try:
            wrap_function_wrapper(
                "crawl4ai.extraction_strategy",
                "LLMExtractionStrategy.extract",
                general_wrap(
                    "crawl4ai.extraction.llm.extract",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.extraction_strategy",
                "LLMExtractionStrategy.run",
                general_wrap(
                    "crawl4ai.extraction.llm.run",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.extraction_strategy",
                "JsonCssExtractionStrategy.extract",
                general_wrap(
                    "crawl4ai.extraction.css.extract",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.extraction_strategy",
                "JsonXPathExtractionStrategy.extract",
                general_wrap(
                    "crawl4ai.extraction.xpath.extract",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.extraction_strategy",
                "CosineStrategy.extract",
                general_wrap(
                    "crawl4ai.extraction.cosine.extract",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.extraction_strategy",
                "RegexExtractionStrategy.extract",
                general_wrap(
                    "crawl4ai.extraction.regex.extract",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        # Wrap Core Processing Operations (0.7.x)
        try:
            wrap_function_wrapper(
                "crawl4ai.async_webcrawler",
                "AsyncWebCrawler.aprocess_html",
                async_general_wrap(
                    "crawl4ai.aprocess_html",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.async_webcrawler",
                "AsyncWebCrawler.aseed_urls",
                async_general_wrap(
                    "crawl4ai.aseed_urls",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        # Wrap Monitor Operations for Business Intelligence
        try:
            wrap_function_wrapper(
                "crawl4ai.components.crawler_monitor",
                "CrawlerMonitor.update_task",
                general_wrap(
                    "crawl4ai.monitor.update_task",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

        try:
            wrap_function_wrapper(
                "crawl4ai.components.crawler_monitor",
                "CrawlerMonitor.get_summary",
                general_wrap(
                    "crawl4ai.monitor.get_summary",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )
        except Exception:
            pass  # Module may not exist in all versions

    def _uninstrument(self, **kwargs):
        """
        Proper uninstrumentation logic to revert patched methods.

        Note: wrapt automatically handles uninstrumentation for
        wrap_function_wrapper calls when the instrumentor is disabled.
        """
        # Uninstrumentation is handled automatically by wrapt
        return



# pylint: disable=useless-return, bad-staticmethod-argument, disable=duplicate-code
"""
Optimized Auto Instrumentation of mem0 Functions following OpenLIT Framework Guide.
"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.mem0.mem0 import mem0_wrap
from openlit.instrumentation.mem0.async_mem0 import async_mem0_wrap

_instruments = ("mem0ai >= 0.1.32",)

# Optimized method configuration with performance priorities
SYNC_METHODS = [
    # Memory initialization
    {
        "package": "mem0",
        "object": "Memory.__init__",
        "endpoint": "memory init",
        "priority": "critical",
    },
    # Core Memory operations
    {
        "package": "mem0",
        "object": "Memory.add",
        "endpoint": "memory add",
        "priority": "high",
    },
    {
        "package": "mem0",
        "object": "Memory.search",
        "endpoint": "memory search",
        "priority": "high",
    },
    {
        "package": "mem0",
        "object": "Memory.get",
        "endpoint": "memory get",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "Memory.get_all",
        "endpoint": "memory get_all",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "Memory.update",
        "endpoint": "memory update",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "Memory.delete",
        "endpoint": "memory delete",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "Memory.delete_all",
        "endpoint": "memory delete_all",
        "priority": "low",
    },
    {
        "package": "mem0",
        "object": "Memory.history",
        "endpoint": "memory history",
        "priority": "low",
    },
    {
        "package": "mem0",
        "object": "Memory.reset",
        "endpoint": "memory reset",
        "priority": "low",
    },
    {
        "package": "mem0",
        "object": "Memory.chat",
        "endpoint": "memory chat",
        "priority": "medium",
    },
    # Internal operations for detailed tracing
    {
        "package": "mem0.memory.main",
        "object": "Memory._add_to_vector_store",
        "endpoint": "memory add_to_vector_store",
        "priority": "medium",
    },
    {
        "package": "mem0.memory.main",
        "object": "Memory._add_to_graph",
        "endpoint": "memory add_to_graph",
        "priority": "medium",
    },
    {
        "package": "mem0.memory.main",
        "object": "Memory._search_vector_store",
        "endpoint": "memory search_vector_store",
        "priority": "medium",
    },
    {
        "package": "mem0.memory.main",
        "object": "Memory._get_all_from_vector_store",
        "endpoint": "memory get_all_from_vector_store",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "Memory._create_memory",
        "endpoint": "memory create_memory",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "Memory._update_memory",
        "endpoint": "memory update_memory",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "Memory._delete_memory",
        "endpoint": "memory delete_memory",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "Memory._create_procedural_memory",
        "endpoint": "memory create_procedural_memory",
        "priority": "low",
    },
]

ASYNC_METHODS = [
    # AsyncMemory initialization
    {
        "package": "mem0",
        "object": "AsyncMemory.__init__",
        "endpoint": "memory init",
        "priority": "critical",
    },
    # Core AsyncMemory operations
    {
        "package": "mem0",
        "object": "AsyncMemory.add",
        "endpoint": "memory add",
        "priority": "high",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.search",
        "endpoint": "memory search",
        "priority": "high",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.get",
        "endpoint": "memory get",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.get_all",
        "endpoint": "memory get_all",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.update",
        "endpoint": "memory update",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.delete",
        "endpoint": "memory delete",
        "priority": "medium",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.delete_all",
        "endpoint": "memory delete_all",
        "priority": "low",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.history",
        "endpoint": "memory history",
        "priority": "low",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.reset",
        "endpoint": "memory reset",
        "priority": "low",
    },
    {
        "package": "mem0",
        "object": "AsyncMemory.chat",
        "endpoint": "memory chat",
        "priority": "medium",
    },
    # Internal async operations for detailed tracing
    {
        "package": "mem0.memory.main",
        "object": "AsyncMemory._add_to_graph",
        "endpoint": "memory add_to_graph",
        "priority": "medium",
    },
    {
        "package": "mem0.memory.main",
        "object": "AsyncMemory._search_vector_store",
        "endpoint": "memory search_vector_store",
        "priority": "medium",
    },
    {
        "package": "mem0.memory.main",
        "object": "AsyncMemory._get_all_from_vector_store",
        "endpoint": "memory get_all_from_vector_store",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "AsyncMemory._create_memory",
        "endpoint": "memory create_memory",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "AsyncMemory._update_memory",
        "endpoint": "memory update_memory",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "AsyncMemory._delete_memory",
        "endpoint": "memory delete_memory",
        "priority": "low",
    },
    {
        "package": "mem0.memory.main",
        "object": "AsyncMemory._create_procedural_memory",
        "endpoint": "memory create_procedural_memory",
        "priority": "low",
    },
]


class Mem0Instrumentor(BaseInstrumentor):
    """Optimized instrumentor for mem0's client library."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        """Optimized instrumentation with performance considerations."""
        # Extract configuration
        application_name = kwargs.get("application_name")
        environment = kwargs.get("environment")
        tracer = kwargs.get("tracer")
        pricing_info = kwargs.get("pricing_info")
        capture_message_content = kwargs.get("capture_message_content")

        # Cache version lookup
        try:
            version = importlib.metadata.version("mem0ai")
        except importlib.metadata.PackageNotFoundError:
            version = "unknown"

        # Pre-create wrapper args
        wrapper_args = (
            version,
            environment,
            application_name,
            tracer,
            pricing_info,
            capture_message_content,
        )

        # Instrument sync methods
        self._wrap_methods(SYNC_METHODS, mem0_wrap, wrapper_args)

        # Instrument async methods
        self._wrap_methods(ASYNC_METHODS, async_mem0_wrap, wrapper_args)

    def _wrap_methods(self, methods, wrapper_func, wrapper_args):
        """Efficiently wrap methods with error handling."""
        for method_config in methods:
            try:
                wrap_function_wrapper(
                    method_config["package"],
                    method_config["object"],
                    wrapper_func(method_config["endpoint"], *wrapper_args),
                )
            except (ImportError, AttributeError):
                # Gracefully handle missing methods in different mem0 versions
                pass

    @staticmethod
    def _uninstrument(self, **kwargs):
        """Uninstrument mem0."""
        # Currently no cleanup needed



# pylint: disable=bare-except, broad-exception-caught
"""
This module has functions to calculate model costs based on tokens and to fetch pricing information.
"""

import os
import json
import logging
from urllib.parse import urlparse
from typing import Any, Dict, List, Tuple
import math
import requests
from opentelemetry.sdk.resources import (
    SERVICE_NAME,
    TELEMETRY_SDK_NAME,
    DEPLOYMENT_ENVIRONMENT,
)
from opentelemetry.trace import Status, StatusCode
from opentelemetry._events import Event
from openlit.semcov import SemanticConvention

# Set up logging
logger = logging.getLogger(__name__)


def response_as_dict(response):
    """
    Return parsed response as a dict
    """

    # pylint: disable=no-else-return
    if isinstance(response, dict):
        return response
    if hasattr(response, "model_dump"):
        return response.model_dump()
    elif hasattr(response, "parse"):
        return response_as_dict(response.parse())
    else:
        return response


def get_env_variable(name, arg_value, error_message):
    """
    Retrieve an environment variable if the argument is not provided
    """

    if arg_value is not None:
        return arg_value
    value = os.getenv(name)
    if not value:
        logging.error(error_message)
        raise RuntimeError(error_message)
    return value


def general_tokens(text):
    """
    Calculate the number of tokens a given text would take up.
    """

    return math.ceil(len(text) / 2)


def get_chat_model_cost(model, pricing_info, prompt_tokens, completion_tokens):
    """
    Retrieve the cost of processing for a given model based on prompt and tokens.
    """

    try:
        cost = ((prompt_tokens / 1000) * pricing_info["chat"][model]["promptPrice"]) + (
            (completion_tokens / 1000) * pricing_info["chat"][model]["completionPrice"]
        )
    except:
        cost = 0
    return cost


def get_embed_model_cost(model, pricing_info, prompt_tokens):
    """
    Retrieve the cost of processing for a given model based on prompt tokens.
    """

    try:
        cost = (prompt_tokens / 1000) * pricing_info["embeddings"][model]
    except:
        cost = 0
    return cost


def get_image_model_cost(model, pricing_info, size, quality):
    """
    Retrieve the cost of processing for a given model based on image size and quailty.
    """

    try:
        cost = pricing_info["images"][model][quality][size]
    except:
        cost = 0
    return cost


def get_audio_model_cost(model, pricing_info, prompt, duration=None):
    """
    Retrieve the cost of processing for a given model based on prompt.
    """

    try:
        if prompt:
            cost = (len(prompt) / 1000) * pricing_info["audio"][model]
        else:
            cost = duration * pricing_info["audio"][model]
    except:
        cost = 0
    return cost


def fetch_pricing_info(pricing_json=None):
    """
    Fetches pricing information from a specified URL or File Path.
    """

    if pricing_json:
        is_url = urlparse(pricing_json).scheme != ""
        if is_url:
            pricing_url = pricing_json
        else:
            try:
                with open(pricing_json, mode="r", encoding="utf-8") as f:
                    return json.load(f)
            except FileNotFoundError:
                logger.error("Pricing information file not found: %s", pricing_json)
            except json.JSONDecodeError:
                logger.error("Error decoding JSON from file: %s", pricing_json)
            except Exception as file_err:
                logger.error(
                    "Unexpected error occurred while reading file: %s", file_err
                )
            return {}
    else:
        pricing_url = (
            "https://raw.githubusercontent.com/openlit/openlit/main/assets/pricing.json"
        )
    try:
        # Set a timeout of 10 seconds for both the connection and the read
        response = requests.get(pricing_url, timeout=20)
        response.raise_for_status()
        return response.json()
    except requests.HTTPError as http_err:
        logger.error("HTTP error occured while fetching pricing info: %s", http_err)
    except Exception as err:
        logger.error("Unexpected error occurred while fetching pricing info: %s", err)
    return {}


def handle_exception(span, e):
    """Handles Exception when LLM Function fails or trace creation fails."""

    span.record_exception(e)
    span.set_status(Status(StatusCode.ERROR))


def calculate_ttft(timestamps: List[float], start_time: float) -> float:
    """
    Calculate the time to the first tokens.
    """

    if timestamps:
        return timestamps[0] - start_time
    return 0.0


def calculate_tbt(timestamps: List[float]) -> float:
    """
    Calculate the average time between tokens.
    """

    if len(timestamps) > 1:
        time_diffs = [
            timestamps[i] - timestamps[i - 1] for i in range(1, len(timestamps))
        ]
        return sum(time_diffs) / len(time_diffs)
    return 0.0


def create_metrics_attributes(
    service_name: str,
    deployment_environment: str,
    operation: str,
    system: str,
    request_model: str,
    server_address: str,
    server_port: int,
    response_model: str,
) -> Dict[Any, Any]:
    """
    Returns OTel metrics attributes
    """

    return {
        TELEMETRY_SDK_NAME: "openlit",
        SERVICE_NAME: service_name,
        DEPLOYMENT_ENVIRONMENT: deployment_environment,
        SemanticConvention.GEN_AI_OPERATION: operation,
        SemanticConvention.GEN_AI_SYSTEM: system,
        SemanticConvention.GEN_AI_REQUEST_MODEL: request_model,
        SemanticConvention.SERVER_ADDRESS: server_address,
        SemanticConvention.SERVER_PORT: server_port,
        SemanticConvention.GEN_AI_RESPONSE_MODEL: response_model,
    }


def set_server_address_and_port(
    client_instance: Any, default_server_address: str, default_server_port: int
) -> Tuple[str, int]:
    """
    Determines and returns the server address and port based on the provided client's `base_url`,
    using defaults if none found or values are None.
    """

    # Try getting base_url from multiple potential attributes
    base_client = getattr(client_instance, "_client", None)
    base_url = getattr(base_client, "base_url", None)

    if not base_url:
        # Attempt to get endpoint from instance._config.endpoint if base_url is not set
        config = getattr(client_instance, "_config", None)
        base_url = getattr(config, "endpoint", None)

    if not base_url:
        # Attempt to get server_url from instance.sdk_configuration.server_url
        config = getattr(client_instance, "sdk_configuration", None)
        base_url = getattr(config, "server_url", None)

    if not base_url:
        # Attempt to get host from instance.config.host (used by Pinecone and other vector DBs)
        config = getattr(client_instance, "config", None)
        base_url = getattr(config, "host", None)

    if base_url:
        if isinstance(base_url, str):
            # Check if it's a full URL or just a hostname
            if base_url.startswith(("http://", "https://")):
                url = urlparse(base_url)
                server_address = url.hostname or default_server_address
                server_port = url.port if url.port is not None else default_server_port
            else:
                # If it's just a hostname (like Pinecone's case), use it directly
                server_address = base_url
                server_port = default_server_port
        else:  # base_url might not be a str; handle as an object.
            server_address = getattr(base_url, "host", None) or default_server_address
            port_attr = getattr(base_url, "port", None)
            server_port = port_attr if port_attr is not None else default_server_port
    else:  # no base_url or endpoint provided; use defaults.
        server_address = default_server_address
        server_port = default_server_port

    return server_address, server_port


def otel_event(name, attributes, body):
    """
    Returns an OpenTelemetry Event object
    """

    return Event(
        name=name,
        attributes=attributes,
        body=body,
    )


def extract_and_format_input(messages):
    """
    Process a list of messages to extract content and categorize
    them into fixed roles like 'user', 'assistant', 'system', 'tool'.
    """

    fixed_roles = ["user", "assistant", "system", "tool", "developer"]
    formatted_messages = {
        role_key: {"role": "", "content": ""} for role_key in fixed_roles
    }

    # Check if input is a simple string
    if isinstance(messages, str):
        formatted_messages["user"] = {"role": "user", "content": messages}
        return formatted_messages

    for message in messages:
        message = response_as_dict(message)

        role = message.get("role")
        if role not in fixed_roles:
            continue

        content = message.get("content", "")

        # Prepare content as a string, handling both list and str
        if isinstance(content, list):
            content_str = ", ".join(str(item) for item in content)
        else:
            content_str = content

        # Set the role in the formatted message and concatenate content
        if not formatted_messages[role]["role"]:
            formatted_messages[role]["role"] = role

        if formatted_messages[role]["content"]:
            formatted_messages[role]["content"] += " " + content_str
        else:
            formatted_messages[role]["content"] = content_str

    return formatted_messages


# To be removed one the change to log events (from span events) is complete
def concatenate_all_contents(formatted_messages):
    """
    Concatenate all 'content' fields into a single strin
    """
    return " ".join(
        message_data["content"]
        for message_data in formatted_messages.values()
        if message_data["content"]
    )


def format_and_concatenate(messages):
    """
    Process a list of messages to extract content, categorize them by role,
    and concatenate all 'content' fields into a single string with role: content format.
    """

    formatted_messages = {}

    # Check if input is a simple string
    if isinstance(messages, str):
        formatted_messages["user"] = {"role": "user", "content": messages}
    elif isinstance(messages, list) and all(isinstance(m, str) for m in messages):
        # If it's a list of strings, each string is 'user' input
        user_content = " ".join(messages)
        formatted_messages["user"] = {"role": "user", "content": user_content}
    else:
        for message in messages:
            message = response_as_dict(message)
            role = message.get(
                "role", "unknown"
            )  # Default to 'unknown' if no role is specified
            content = message.get("content", "")

            # Initialize role in formatted messages if not present
            if role not in formatted_messages:
                formatted_messages[role] = {"role": role, "content": ""}

            # Handle list of dictionaries in content
            if isinstance(content, list):
                content_str = []
                for item in content:
                    if isinstance(item, dict):
                        # Collect text or other attributes as needed
                        text = item.get("text", "")
                        image_url = item.get("image_url", "")
                        content_str.append(text)
                        content_str.append(image_url)
                content_str = ", ".join(filter(None, content_str))
            else:
                content_str = content

            # Concatenate content
            if formatted_messages[role]["content"]:
                formatted_messages[role]["content"] += " " + content_str
            else:
                formatted_messages[role]["content"] = content_str

    # Concatenate role and content for all messages
    return " ".join(
        f"{message_data['role']}: {message_data['content']}"
        for message_data in formatted_messages.values()
        if message_data["content"]
    )


def common_span_attributes(
    scope,
    gen_ai_operation,
    gen_ai_system,
    server_address,
    server_port,
    request_model,
    response_model,
    environment,
    application_name,
    is_stream,
    tbt,
    ttft,
    version,
):
    """
    Set common span attributes for both chat and RAG operations.
    """

    scope._span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
    scope._span.set_attribute(SemanticConvention.GEN_AI_OPERATION, gen_ai_operation)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SYSTEM, gen_ai_system)
    scope._span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
    scope._span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL, request_model)
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL, response_model)
    scope._span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    scope._span.set_attribute(SERVICE_NAME, application_name)
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_IS_STREAM, is_stream)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SERVER_TBT, tbt)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SERVER_TTFT, ttft)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION, version)


def record_completion_metrics(
    metrics,
    gen_ai_operation,
    gen_ai_system,
    server_address,
    server_port,
    request_model,
    response_model,
    environment,
    application_name,
    start_time,
    end_time,
    cost,
    input_tokens,
    output_tokens,
    tbt,
    ttft,
):
    """
    Record completion metrics for the operation.
    """

    attributes = create_metrics_attributes(
        operation=gen_ai_operation,
        system=gen_ai_system,
        server_address=server_address,
        server_port=server_port,
        request_model=request_model,
        response_model=response_model,
        service_name=application_name,
        deployment_environment=environment,
    )
    metrics["genai_client_usage_tokens"].record(
        input_tokens + output_tokens, attributes
    )
    metrics["genai_client_operation_duration"].record(end_time - start_time, attributes)
    metrics["genai_server_tbt"].record(tbt, attributes)
    metrics["genai_server_ttft"].record(ttft, attributes)
    metrics["genai_requests"].add(1, attributes)
    metrics["genai_completion_tokens"].add(output_tokens, attributes)
    metrics["genai_prompt_tokens"].add(input_tokens, attributes)
    metrics["genai_cost"].record(cost, attributes)


def record_embedding_metrics(
    metrics,
    gen_ai_operation,
    gen_ai_system,
    server_address,
    server_port,
    request_model,
    response_model,
    environment,
    application_name,
    start_time,
    end_time,
    input_tokens,
    cost,
):
    """
    Record embedding-specific metrics for the operation.
    """

    attributes = create_metrics_attributes(
        operation=gen_ai_operation,
        system=gen_ai_system,
        server_address=server_address,
        server_port=server_port,
        request_model=request_model,
        response_model=response_model,
        service_name=application_name,
        deployment_environment=environment,
    )
    metrics["genai_client_usage_tokens"].record(input_tokens, attributes)
    metrics["genai_client_operation_duration"].record(end_time - start_time, attributes)
    metrics["genai_requests"].add(1, attributes)
    metrics["genai_prompt_tokens"].add(input_tokens, attributes)
    metrics["genai_cost"].record(cost, attributes)


def record_audio_metrics(
    metrics,
    gen_ai_operation,
    gen_ai_system,
    server_address,
    server_port,
    request_model,
    response_model,
    environment,
    application_name,
    start_time,
    end_time,
    cost,
):
    """
    Record audio-specific metrics for the operation.
    """

    attributes = create_metrics_attributes(
        operation=gen_ai_operation,
        system=gen_ai_system,
        server_address=server_address,
        server_port=server_port,
        request_model=request_model,
        response_model=response_model,
        service_name=application_name,
        deployment_environment=environment,
    )
    metrics["genai_client_operation_duration"].record(end_time - start_time, attributes)
    metrics["genai_requests"].add(1, attributes)
    metrics["genai_cost"].record(cost, attributes)


def record_image_metrics(
    metrics,
    gen_ai_operation,
    gen_ai_system,
    server_address,
    server_port,
    request_model,
    response_model,
    environment,
    application_name,
    start_time,
    end_time,
    cost,
):
    """
    Record image-specific metrics for the operation.
    """

    attributes = create_metrics_attributes(
        operation=gen_ai_operation,
        system=gen_ai_system,
        server_address=server_address,
        server_port=server_port,
        request_model=request_model,
        response_model=response_model,
        service_name=application_name,
        deployment_environment=environment,
    )
    metrics["genai_client_operation_duration"].record(end_time - start_time, attributes)
    metrics["genai_requests"].add(1, attributes)
    metrics["genai_cost"].record(cost, attributes)


def common_db_span_attributes(
    scope,
    db_system,
    server_address,
    server_port,
    environment,
    application_name,
    version,
):
    """
    Set common span attributes for database operations.
    """

    scope._span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_OPERATION,
        SemanticConvention.GEN_AI_OPERATION_TYPE_VECTORDB,
    )
    scope._span.set_attribute(SemanticConvention.DB_SYSTEM_NAME, db_system)
    scope._span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
    scope._span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
    scope._span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    scope._span.set_attribute(SERVICE_NAME, application_name)
    scope._span.set_attribute(SemanticConvention.DB_SDK_VERSION, version)


def common_framework_span_attributes(
    scope,
    framework_system,
    server_address,
    server_port,
    environment,
    application_name,
    version,
    endpoint,
    instance=None,
):
    """
    Set common span attributes for GenAI framework operations.
    """

    scope._span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
    scope._span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION, version)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SYSTEM, framework_system)
    scope._span.set_attribute(SemanticConvention.GEN_AI_OPERATION, endpoint)
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_MODEL,
        getattr(instance, "model_name", "unknown") if instance else "unknown",
    )
    scope._span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
    scope._span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
    scope._span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    scope._span.set_attribute(SERVICE_NAME, application_name)
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_OPERATION_DURATION,
        scope._end_time - scope._start_time,
    )


def record_framework_metrics(
    metrics,
    gen_ai_operation,
    gen_ai_system,
    server_address,
    server_port,
    environment,
    application_name,
    start_time,
    end_time,
):
    """
    Record basic framework metrics for the operation (only gen_ai_requests counter).
    """

    attributes = create_metrics_attributes(
        operation=gen_ai_operation,
        system=gen_ai_system,
        server_address=server_address,
        server_port=server_port,
        request_model="unknown",
        response_model="unknown",
        service_name=application_name,
        deployment_environment=environment,
    )
    metrics["genai_requests"].add(1, attributes)
    metrics["genai_client_operation_duration"].record(end_time - start_time, attributes)


def record_db_metrics(
    metrics,
    db_system,
    server_address,
    server_port,
    environment,
    application_name,
    start_time,
    end_time,
):
    """
    Record database-specific metrics for the operation.
    """

    attributes = create_metrics_attributes(
        operation=SemanticConvention.GEN_AI_OPERATION_TYPE_VECTORDB,
        system=db_system,
        request_model=db_system,
        server_address=server_address,
        server_port=server_port,
        response_model=db_system,
        service_name=application_name,
        deployment_environment=environment,
    )
    metrics["db_requests"].add(1, attributes)
    metrics["db_client_operation_duration"].record(end_time - start_time, attributes)
